#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np
from scipy import linalg as spl
from numpy import linalg as npl


def get_degeneracies ( E_k, bnd ):

  all_degen = []

  E_k_round = np.around(E_k, decimals=5)
  
  for ispin in range(E_k_round.shape[2]):

    by_spin = []
    for ik in range(E_k_round.shape[0]):
      by_kp = []
      eV = np.unique(E_k_round[ik,:,ispin][:-1][np.isclose(E_k_round[ik,:,ispin][1:],
                                                           E_k_round[ik,:,ispin][:-1],
                                                           atol=1.e-5)])

      for i in range(len(eV)):
        inds = np.where(np.isclose(E_k_round[ik,:,ispin],eV[i],atol=1.e-5))[0]

        if len(inds) > 1 and np.all(inds < bnd):                    
          by_kp.append(inds)

      by_spin.append(by_kp)

    all_degen.append(by_spin)

  return all_degen


def do_pao_eigh ( data_controller ):
  from .communication import gather_scatter
  from numpy.linalg import eigh
  from mpi4py import MPI

  rank = MPI.COMM_WORLD.Get_rank()

  arrays,attributes = data_controller.data_dicts()

  snktot,nawf,_,nspin = arrays['Hksp'].shape

  arrays['E_k'] = np.zeros((snktot,nawf,nspin), dtype=float)
  arrays['v_k'] = np.zeros((snktot,nawf,nawf,nspin), dtype=complex)

  for ispin in range(nspin):
    for n in range(snktot):
      arrays['E_k'][n,:,ispin],arrays['v_k'][n,:,:,ispin] = eigh(arrays['Hksp'][n,:,:,ispin], UPLO='U')

  # arrays['degen'] = get_degeneracies(arrays['E_k'], attributes['nawf'])
  arrays['degen'] = get_degeneracies(arrays['E_k'], attributes['bnd'])


def do_eigh_calc ( HRaux, SRaux, kq, R, read_S ):

  # Compute bands on a selected mesh in the BZ

  nkpi = kq.shape[0]
  nawf = HRaux.shape[0]
  nspin = HRaux.shape[-1]

  Hks_int = band_loop_H(HRaux, kq, R)

  if read_S:
    Sks_int = band_loop_S(SRaux, kq, R)

  E_kp = np.empty((nkpi,nawf,nspin), dtype=float)
  v_kp = np.empty((nkpi,nawf,nawf,nspin), dtype=complex)

  for ispin in range(nspin):
    for ik in range(nkpi):
      if read_S:
        E_kp[ik,:,ispin],v_kp[ik,:,:,ispin] = spl.eigh(Hks_int[:,:,ik,ispin],Sks_int[:,:,ik])
      else:
        E_kp[ik,:,ispin],v_kp[ik,:,:,ispin] = npl.eigh(Hks_int[:,:,ik,ispin],UPLO='U')

  return (E_kp, v_kp)


### R_wght assumed to be 1
def band_loop_H ( HRaux, kq, R ):

  nkpi = kq.shape[0]
  nawf,_,nk1,nk2,nk3,nspin = HRaux.shape
  auxh = np.empty((nawf,nawf,nkpi,nspin), dtype=complex)
  HRaux = np.reshape(HRaux, (nawf,nawf,nk1*nk2*nk3,nspin), order='C')

  for ik in range(nkpi):
    for ispin in range(nspin):
       auxh[:,:,ik,ispin] = np.sum(HRaux[:,:,:,ispin]*np.exp(2.j*np.pi*kq[ik,:].dot(R[:,:].T)), axis=2)

  return auxh


def band_loop_S ( SRaux, kq, R ):

  nkpi = kq.shape[0]
  nawf,_,nk1,nk2,nk3 = SRaux.shape
  auxs = np.empty((nawf,nawf,nkpi), dtype=complex)
  SRaux = np.reshape(SRaux, (nawf,nawf,nk1*nk2*nk3), order='C')

  for ik in range(nkpi):
    auxs[:,:,ik] = np.sum(SRaux[:,:,:]*np.exp(2.j*np.pi*kq[ik,:].dot(R[:,:].T)), axis=2)

  return auxs



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .


class ErrorHandler:

  def __init__ ( self ):
    pass

  def report_exception ( self, mname='UNKNOWN' ):
    '''
    Print the exception and traceback. Print suggestion if the error is recognized as a user mistake.

    Arguments:
        mname (str): Module name (name of function in PAOFLOW)

    Returns:
        None
    '''
    import sys
    import traceback
    from .defs.module_prerequisites import key_error_strings,report_pre_reqs,module_pre_reqs

    etype, evalue, etb = sys.exc_info()
    print('Exception: ', etype, evalue, flush=True)
    traceback.print_tb(etb)

    if etype is KeyError:

      if mname in module_pre_reqs:
        print('HHH',mname, type(mname))
        pre_reqs = module_pre_reqs[mname]
        if len(pre_reqs) > 1:
          pr_str = ', '.join(pre_reqs[:-1]) + ' and %s'%pre_reqs[-1]
        else:
          pr_str = pre_reqs[0]

        print('')
        print(report_pre_reqs%(pr_str,mname), flush=True)

      if evalue.args[0] in key_error_strings:
        print('')
        print('SUGGESTION: %s\n'%key_error_strings[evalue.args[0]], flush=True)



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import xml.etree.cElementTree as ET
from mpi4py import MPI
import numpy as np
import spglib

def parse_vasprun_data ( data_controller, fname, symprec = 1e-4 ):
  '''
  Parse the data_file_schema.xml file produced by Quantum Espresso.
  Populated the DataController object with all necessary information.

  Arguments:
    data_controller (DataController): Data controller to populate
    fname (str): Path and name of the xml file.
  '''
  import re

  arry,attr = data_controller.data_dicts()
  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  verbose = attr['verbose']
  Ryd2eV = 13.605826  # Note this number is slightly different between VASP and QE
  AUTOA = 0.529177249  # see vasp/src/constant.inc

  tree = ET.parse(fname)
  root = tree.getroot()
  
  version = root.find("./generator/i[@name='version']").text
  if verbose: print('VASP version: {0:s}'.format(version))
  qe_version = list(map(int, version.strip().split(".")))

  '''read parameters'''
  parameters = root.find('parameters')
  lsorbit = parameters.find(".//i[@name='LSORBIT']").text.strip()
  dftSO = True if lsorbit == 'T' else False
  nelec = float(parameters.find(".//i[@name='NELECT']").text.strip())
  nspin = int(parameters.find(".//i[@name='ISPIN']").text.strip())
  nbnds = int(parameters.find(".//i[@name='NBANDS']").text.strip())
  nawf = nbnds
  ecutwfc = float(parameters.find(".//i[@name='ENMAX']").text.strip())/Ryd2eV # convert to Ryd
  ecutrho = ecutwfc * 5
  '''end'''

  # tag = atominfo  
  atom_elem = root.find('atominfo')
  species,pseudos = [],[]
  specs = atom_elem.find("./array[@name='atomtypes']/set")
  for i,s in enumerate(specs.findall(".//c")):
    if i%5 == 1:
      species.append(s.text.strip())
    elif i%5 == 4:
      pseudos.append(s.text.strip())  # Reference only, pseudos not used

  alat = 1.0
  natoms = int(atom_elem.find("atoms").text.strip())
  atms = atom_elem.find("./array[@name='atoms']/set")
  atoms = []
  for i,a in enumerate(atms.findall(".//c")):
    if i%2 == 0:
      atoms.append(a.text.strip())

  # tag = structure
  struct_elem = root.find("./structure[@name='finalpos']")
  basis = struct_elem.find("./crystal/varray[@name='basis']")
  rec_basis = struct_elem.find("./crystal/varray[@name='rec_basis']")
  a_Angstrom, b_Angstrom = [], []
  for a_vec in basis.findall("./v"):
    a_Angstrom.append([float(a) for a in a_vec.text.strip().split()])
  for b_vec in rec_basis.findall("./v"):
    b_Angstrom.append([float(b) for b in b_vec.text.strip().split()])
  # The definitions here are different from read_QE_xml.py
  # We fix alat = 1.0 Bohr for VASP structures.
  a_vectors = np.array(a_Angstrom)/AUTOA # Convert Angstrom to Bohr
  b_vectors = np.array(b_Angstrom)*AUTOA
  omega = alat ** 3 * a_vectors[0, :].dot(np.cross(a_vectors[1, :], a_vectors[2, :]))

  pos_list = []
  positions = struct_elem.find("./varray[@name='positions']")
  for x in positions.findall("./v"):
    pos_list.append([float(t) for t in x.text.strip().split()])
  pos_arry = np.array(pos_list)
  tau = pos_arry @ a_vectors  # Atomic coord in cartesian, Bohr

  # tag = kpoints
  k_elem = root.find("./kpoints")
  kmesh = k_elem.find(".//v[@name='divisions']").text.strip().split()
  nk1, nk2, nk3 = int(kmesh[0]), int(kmesh[1]), int(kmesh[2])
  # if Monkhorst-Pack grid with even nk, shift of 0.5
  # otherwise, no shift
  mpgrid = k_elem.find("./generation").attrib['param']
  if mpgrid[0] == "M":
    k1, k2, k3 = int(nk1%2==0), int(nk2%2==0), int(nk3%2==0)
  else:
    k1, k2, k3 = 0, 0, 0
  if verbose: print('Monkhorst and Pack grid:', nk1, nk2, nk3, k1, k2, k3)

  kpnts_temp, kpnt_weights_temp = [], []
  for ks in k_elem.findall("./varray[@name='kpointlist']/v"):
    kpnts_temp.append([float(k) for k in ks.text.strip().split()])
  kpnts = np.array(kpnts_temp)@b_vectors
  nkpnts = int(kpnts.shape[0])

  for k in k_elem.findall("./varray[@name='weights']/v"):
    kpnt_weights_temp.append(float(k.text.strip()))
  kpnt_weights = np.array(kpnt_weights_temp)

  # Read eigenvalues
  eigs = np.empty((nbnds,nkpnts,nspin), dtype=float)
  E_elem = root.find("./calculation/eigenvalues")
  Efermi = float(root.find(".//dos/i[@name='efermi']").text.strip())
  for i, kpt in enumerate(E_elem.findall(".//set[@comment='spin 1']/set")):
    energy_at_k = []
    for e in kpt.findall("./r"):
      energy_at_k.append(float(e.text.strip().split()[0]))
    eigs[:,i,0] = energy_at_k
  if nspin == 2:
    for i, kpt in enumerate(E_elem.findall(".//set[@comment='spin 2']/set")):
      energy_at_k = []
      for e in kpt.findall("./r"):
        energy_at_k.append(float(e.text.strip().split()[0]))
      eigs[:,i,1] = energy_at_k
  eigs -= Efermi

  # Check metal or insulator
  # Metal if CBM and VBM are less than 0.05 eV apart
  cbm = np.min(eigs[eigs > 0])
  vbm = np.max(eigs[eigs < 0])
  insulator = (cbm-vbm) > 0.05

  # read MAGMOM
  # Note: the magnetic moments in vasprun.xml are the values defined in INCAR
  # The default is ferromagnetic
  magnetization = root.find(".//v[@name='MAGMOM']").text.strip().split()
  magmom = np.asarray([float(m) for m in magnetization])
  if dftSO:
    magmom = magmom.reshape((-1,3))
  finite_magmom = np.any(np.abs(magmom)>1e-3)
  dftMag = (nspin == 1 and dftSO and finite_magmom) or (nspin ==2 and finite_magmom)


  _, atom_numbers = np.unique(atoms, return_inverse=True)
  cell = (a_vectors, pos_arry, atom_numbers)
  if nkpnts == nk1 * nk2 * nk3:
    # Check whether VASP calculation uses symmetry (ISYM = -1, 0, or 2)
    ID = np.identity(3,dtype=int)
    sym_rot_transpose = ID[np.newaxis, :]
    shifts = np.zeros((1, 3))
    sym_info = np.asarray([None])
    time_rev = np.asarray([False])
    eq_atoms = np.arange(natoms)
    eq_atoms = eq_atoms[np.newaxis,:]

  else:
    #  get symmetry from spglib
    if dftMag:
      cell_mag = (a_vectors, pos_arry, atom_numbers, magmom)
      spglib_sym = spglib.get_magnetic_symmetry(cell_mag,symprec=symprec)
      sym_rot = spglib_sym['rotations']
      shifts = spglib_sym['translations']
      time_rev = spglib_sym['time_reversals']
      if dftSO:
        sym_rot = sym_rot[~time_rev,:,:]
        shifts = shifts[~time_rev,:]
        time_rev = np.asarray([False] * sym_rot.shape[0])
    else:
      spglib_sym = spglib.get_symmetry(cell,symprec=symprec)
      sym_rot = spglib_sym['rotations']
      shifts = spglib_sym['translations']
      time_rev = np.asarray([False] * sym_rot.shape[0])


    # Find equivalent atoms
    # vasp/src/symlib.F SUBROUTINE POSMAP

    nops = sym_rot.shape[0]
    eq_atoms = np.zeros((nops, natoms), dtype=int)
    sym_rot_transpose = np.zeros((nops,3,3),dtype=int)
    for i_ops in range(nops):
      ops = sym_rot[i_ops,:,:]
      sym_rot_transpose[i_ops,:,:] = ops.T
      new_pos_arry = pos_arry @ ops.T + shifts[i_ops,:]
      for i_atm in range(natoms):
        for j_atm in range(natoms):
          diff_pos = new_pos_arry[j_atm,:] - pos_arry[i_atm, :]
          if np.linalg.norm(np.mod(diff_pos + 0.5, 1) - 0.5) < symprec:
            eq_atoms[i_ops, j_atm] = i_atm

    if not np.all(np.sum(eq_atoms,axis=1) == natoms*(natoms-1)/2):
      raise Exception("Fail to find equivalent atoms")

    sym_info = np.asarray([None]*nops)

  space_group = spglib.get_spacegroup(cell,symprec=symprec)
  if rank == 0 :
    print('Space group: {0}'.format(space_group))
    print('Insulator: {0}'.format(insulator))
    print('Magnetic: {0}'.format(dftMag))
    print('Number of kpoints: {0:d}'.format(nkpnts))
    print('Number of electrons: {0:f}'.format(nelec))
    print('Number of bands: {0:d}'.format(nbnds))

  attrs = [('qe_version',qe_version),('alat',alat),('nk1',nk1),('nk2',nk2),('nk3',nk3),('ok1',k1),('ok2',k2),('ok3',k3),('natoms',natoms),('ecutrho',ecutrho),('ecutwfc',ecutwfc),('nawf',nawf),('nbnds',nbnds),('nspin',nspin),('nkpnts',nkpnts),('nelec',nelec),('Efermi',Efermi),('omega',omega),('dftSO',dftSO),('dftMAG',dftMag),('insulator',insulator)]
  for s,v in attrs:
    attr[s] = v

  spec = [(species[i],pseudos[i]) for i in range(len(species))]
  arrys = [('tau',tau),('atoms',atoms),('species',spec),('a_vectors',a_vectors),('b_vectors',b_vectors),('equiv_atom',eq_atoms),('kpnts',kpnts),('kpnts_wght',kpnt_weights),('my_eigsmat',eigs),('sym_rot',sym_rot_transpose),('sym_shift',shifts),('sym_info',sym_info),('sym_TR',time_rev)]
  for s,v in arrys:
    arry[s] = v





#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

def get_K_grid_fft ( data_controller ):
  import numpy as np

  arrays,attributes = data_controller.data_dicts()

  nk1 = attributes['nk1']
  nk2 = attributes['nk2']
  nk3 = attributes['nk3']
  b_vectors = arrays['b_vectors']

  nktot = nk1*nk2*nk3
  arrays['kq_wght'] = np.ones((nktot), dtype=float)/nktot
#  return

### Not used
  Kint = np.zeros((3,nktot), dtype=float)
  idk = np.zeros((nk1,nk2,nk3), dtype=int)

  for i in range(nk1):
    for j in range(nk2):
      for k in range(nk3):
        n = k + j*nk3 + i*nk2*nk3
        Rx = float(i)/float(nk1)
        Ry = float(j)/float(nk2)
        Rz = float(k)/float(nk3)
        if Rx >= 0.5: Rx=Rx-1.0
        if Ry >= 0.5: Ry=Ry-1.0
        if Rz >= 0.5: Rz=Rz-1.0
        Rx -= int(Rx)
        Ry -= int(Ry)
        Rz -= int(Rz)
        idk[i,j,k]=n
        Kint[:,n] = Rx*b_vectors[0,:]+Ry*b_vectors[1,:]+Rz*b_vectors[2,:]
  arrays['kgrid'] = Kint
  return

def get_K_grid_fft_crystal ( nk1,nk2,nk3 ):
  import numpy as np

  nktot = nk1*nk2*nk3

  Kint = np.zeros((nktot,3), dtype=float)

  for i in range(nk1):
    for j in range(nk2):
      for k in range(nk3):
        n = k + j*nk3 + i*nk2*nk3
        Rx = float(i)/float(nk1)
        Ry = float(j)/float(nk2)
        Rz = float(k)/float(nk3)
        if Rx >= 0.5: Rx=Rx-1.0
        if Ry >= 0.5: Ry=Ry-1.0
        if Rz >= 0.5: Rz=Rz-1.0
        Rx -= int(Rx)
        Ry -= int(Ry)
        Rz -= int(Rz)

        Kint[n] = Rx,Ry,Rz

  return Kint




#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import os
import numpy as np
from numpy import linalg as npl

# Compute Z2 invariant and topological properties on a selected path in the BZ
def do_topology ( data_controller ):
  from mpi4py import MPI
  from scipy.fftpack import fftshift
  from .constants import LL, ANGSTROM_AU
  from .get_R_grid_fft import get_R_grid_fft
  from .communication import scatter_full,gather_full
  from .kpnts_interpolation_mesh import kpnts_interpolation_mesh

  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  arrays,attributes = data_controller.data_dicts()

  npool = attributes['npool']

  if 'kq' not in arrays:
    kpnts_interpolation_mesh(data_controller)

  HRs = arrays['HRs']
  bnd = attributes['bnd']
  nkpi = arrays['kq'].shape[1]
  nawf,_,nk1,nk2,nk3,nspin = HRs.shape

  if 'Rfft' not in arrays:
    get_R_grid_fft(data_controller, nk1, nk2, nk3)

  ipol = attributes['ipol']
  jpol = attributes['jpol']
  spol = attributes['spol']

  Berry = attributes['Berry']
  eff_mass = attributes['eff_mass']
  spin_Hall = attributes['spin_Hall']

  alat = attributes['alat'] / ANGSTROM_AU
  b_vectors = arrays['b_vectors']

  # Compute Z2 according to Fu, Kane and Mele (2007)
  # Define TRIM points in 2(0-3)/3D(0-7)
  if nspin == 1 and spin_Hall:
    from .pfaffian import pfaffian
    from .do_eigh import do_eigh_calc
    from .clebsch_gordan import clebsch_gordan

    nktrim = 16
    ktrim = np.zeros((nktrim,3),dtype=float)
    ktrim[0] = np.zeros(3,dtype=float)                                  #0 0 0 0
    ktrim[1] = b_vectors[0,:]/2.0                                       #1 1 0 0
    ktrim[2] = b_vectors[1,:]/2.0                                       #2 0 1 0
    ktrim[3] = b_vectors[0,:]/2.0+b_vectors[1,:]/2.0                    #3 1 1 0
    ktrim[4] = b_vectors[2,:]/2.0                                       #4 0 0 1
    ktrim[5] = b_vectors[1,:]/2.0+b_vectors[2,:]/2.0                    #5 0 1 1
    ktrim[6] = b_vectors[2,:]/2.0+b_vectors[0,:]/2.0                    #6 1 0 1
    ktrim[7] = b_vectors[0,:]/2.0+b_vectors[1,:]/2.0+b_vectors[2,:]/2.0 #7 1 1 1
    ktrim[8:16] = -ktrim[:8]

    # Compute eigenfunctions at the TRIM points
    SRs = None
    acbn0 = False
    if 'SRs' in arrays:
      SRs = arrays['SRs']
      acbn0 = True
    E_ktrim,v_ktrim = do_eigh_calc(HRs, SRs, ktrim, arrays['R'], acbn0)

    # Define time reversal operator
    if 'adhoc_SO' in attributes and attributes['adhoc_SO'] == True:
      Sjy = np.zeros((nawf,nawf), dtype=complex)
      sP = 0.5*np.array([[0.0,-1.0j],[1.0j,0.0]])
      for i in range(nawf//2):
        Sjy[i,i] = sP[0,0]
        Sjy[i,i+1] = sP[0,1]
      for i in range(nawf//2, nawf):
        Sjy[i,i-1] = sP[1,0]
        Sjy[i,i] = sP[1,1]
      theta = -1.0j*Sjy
    else:
      theta = -1.0j*clebsch_gordan(nawf, arrays['sh_l'], arrays['sh_j'], 1)

    nkt = nktrim // 2

    nelec = attributes['nelec']
    wl = np.zeros((nkt,nawf,nawf), dtype=complex)
    for ik in range(nkt):
      wl[ik,:,:] = np.conj(v_ktrim[ik,:,:,0].T).dot(theta).dot(np.conj(v_ktrim[ik+nkt,:,:,0]))
      wl[ik,:,:] = wl[ik,:,:] - wl[ik,:,:].T  # enforce skew symmetry
    delta_ik = np.zeros(nkt, dtype=complex)
    for ik in range(nkt):
      delta_ik[ik] = pfaffian(wl[ik,:nelec,:nelec])/np.sqrt(npl.det(wl[ik,:nelec,:nelec]))

    # Write 'Z2.dim'
    with open(os.path.join(attributes['opath'],'Z2'+'.dim'), 'w') as f:
      p2D = np.real(np.prod(delta_ik[:4]))
      v0 = (1 if p2D+1. < 1.e-5 else 0)
      f.write('2D case: v0 = %1d \n' %(v0))

      p3D = np.real(np.prod(delta_ik))
      v0 = (1 if p3D+1. < 1.e-5 else 0)

      p3D = delta_ik[1]*delta_ik[3]*delta_ik[6]*delta_ik[7]
      v1 = (1 if p3D+1. < 1.e-5 else 0)

      p3D = delta_ik[2]*delta_ik[3]*delta_ik[5]*delta_ik[7]
      v2 = (1 if p3D+1. < 1.e-5 else 0)

      p3D = delta_ik[4]*delta_ik[6]*delta_ik[5]*delta_ik[7]
      v3 = (1 if p3D+1. < 1.e-5 else 0)

      f.write('3D case: v0;v1,v2,v3 = %1d;%1d,%1d,%1d \n' %(v0,v1,v2,v3))

  # Compute momenta and kinetic energy
  kq_aux = scatter_full(arrays['kq'].T, npool)
  kq_aux = kq_aux.T

  # Compute R*H(R)
  Rfft = np.reshape(arrays['Rfft'], (nk1*nk2*nk3,3), order='C')
  HRs = np.reshape(HRs, (nawf,nawf,nk1*nk2*nk3,nspin), order='C')
  HRs = np.moveaxis(HRs, 2, 0)
  
  # Compute H(k) on the path
  Hks_aux  = band_loop_H(np.moveaxis(HRs,0,2), Rfft, kq_aux, nawf, nspin)

  HRs_aux = scatter_full(HRs, npool)
  Rfft_aux = scatter_full(Rfft, npool)

  HRs = np.reshape(np.moveaxis(HRs,0,2), (nawf,nawf,nk1,nk2,nk3,nspin), order='C')

  if spin_Hall:
    Sj = arrays['Sj']
    jks = np.zeros((kq_aux.shape[1],3,bnd,bnd,nspin), dtype=complex)

  pks = np.zeros((kq_aux.shape[1],3,bnd,bnd,nspin), dtype=complex)
  for l in range(3):
    dHRs  = np.zeros((HRs_aux.shape[0],nawf,nawf,nspin),dtype=complex)
    for ispin in range(nspin):
      for n in range(nawf):
        for m in range(nawf):
          dHRs[:,n,m,ispin] = 1.0j*alat*ANGSTROM_AU*Rfft_aux[:,l]*HRs_aux[:,n,m,ispin]
    dHRs = gather_full(dHRs, npool)
    if rank != 0:
      dHRs = np.zeros((nk1*nk2*nk3,nawf,nawf,nspin), dtype=complex)
    comm.Bcast(dHRs)
    dHRs = np.moveaxis(dHRs,0,2)

    # Compute dH(k)/dk on the path
    dHks_aux = band_loop_H(dHRs, Rfft, kq_aux, nawf, nspin)

    dHRs = None
    # including correction Dnm
    for ik in range(dHks_aux.shape[0]):
      for ispin in range(nspin):
        dHks_aux[ik,:,:,ispin] = dHks_aux[ik,:,:,ispin] + 1.0j*Hks_aux[ik,:,:,ispin]*arrays['Dnm'][:,:,l]  

    # Compute momenta
    for ik in range(dHks_aux.shape[0]):
      for ispin in range(nspin):
        pks[ik,l,:,:,ispin] = np.conj(arrays['v_k'][ik,:,:,ispin].T).dot(dHks_aux[ik,:,:,ispin]).dot(arrays['v_k'][ik,:,:,ispin])[:bnd,:bnd]

    if spin_Hall:
      for ik in range(pks.shape[0]):
        for ispin in range(nspin):
          jks[ik,l,:,:,ispin] = (np.conj(arrays['v_k'][ik,:,:,ispin].T).dot \
            (0.5*(np.dot(Sj[spol],dHks_aux[ik,:,:,ispin])+np.dot(dHks_aux[ik,:,:,ispin],Sj[spol]))).dot(arrays['v_k'][ik,:,:,ispin]))[:bnd,:bnd]

  if eff_mass == True:
    tks = np.zeros((kq_aux.shape[1],3,3,bnd,bnd,nspin), dtype=complex)

    for l in range(3):
      for lp in range(3):
        d2HRs = np.zeros((HRs_aux.shape[0],nawf,nawf,nspin), dtype=complex)
        for ispin in range(nspin):
          for n in range(nawf):
            for m in range(nawf):
              d2HRs[:,n,m,ispin] = -1.0*alat**2*ANGSTROM_AU**2*Rfft_aux[:,l]*Rfft_aux[:,lp]*HRs_aux[:,n,m,ispin]

        d2HRs = gather_full(d2HRs, npool)
        if rank != 0:
          d2HRs = np.zeros((nk1*nk2*nk3,nawf,nawf,nspin), dtype=complex)
        comm.Bcast(d2HRs)
        d2HRs = np.moveaxis(d2HRs, 0, 2)

        # Compute d2H(k)/dk*dkp on the path
        d2Hks_aux = band_loop_H(d2HRs, Rfft, kq_aux, nawf, nspin)

        d2HRs = None

        # Compute kinetic energy
        for ik in range(d2Hks_aux.shape[0]):
          for ispin in range(nspin):
            tks[ik,l,lp,:,:,ispin] = (np.conj(arrays['v_k'][ik,:,:,ispin].T).dot(d2Hks_aux[ik,:,:,ispin]).dot(arrays['v_k'][ik,:,:,ispin]))[:bnd,:bnd]

        d2Hks_aux = None

    # Compute effective mass
    mkm1 = np.zeros((tks.shape[0],bnd,3,3,nspin), dtype=complex)
    for ik in range(tks.shape[0]):
      for ispin in range(nspin):
        for n in range(bnd):
          for m in range(bnd):
            if m != n:
              mkm1[ik,n,ipol,jpol,ispin] += (pks[ik,ipol,n,m,ispin]*pks[ik,jpol,m,n,ispin]+pks[ik,jpol,n,m,ispin]*pks[ik,ipol,m,n,ispin]) / \
                            (arrays['E_k'][ik,n,ispin]-arrays['E_k'][ik,m,ispin]+1.e-16)
            else:
              mkm1[ik,n,ipol,jpol,ispin] += tks[ik,ipol,jpol,n,n,ispin]


    tks = None

    mkm1 = gather_full(mkm1, npool)

#### Write to data_controller
    #mkm1 *= ELECTRONVOLT_SI**2/H_OVER_TPI**2*ELECTRONMASS_SI
    if rank == 0:
      for ispin in range(nspin):
        f = open(os.path.join(attributes['opath'],'effmass'+'_'+str(LL[ipol])+str(LL[jpol])+'_'+str(ispin)+'.dat'),'w')
        for ik in range(nkpi):
          s="%d\t"%ik
          for  j in np.real(mkm1[ik,:bnd,ipol,jpol,ispin]):s += "% 3.5f\t"%j
          s+="\n"
          f.write(s)
        f.close()

    mkm1 = None

  HRs_aux = None
  HRs = None

  # Compute Berry curvature
  if Berry or spin_Hall:
    deltab = 0.05
    mu = -0.2 # chemical potential in eV)
    Om_zk = np.zeros((pks.shape[0],1), dtype=float)
    Om_znk = np.zeros((pks.shape[0],bnd), dtype=float)
    Omj_zk = (np.zeros((pks.shape[0],1), dtype=float) if spin_Hall else None)
    Omj_znk = (np.zeros((pks.shape[0],bnd), dtype=float) if spin_Hall else None)
    for ik in range(pks.shape[0]):
      for n in range(bnd):
        for m in range(bnd):
          if m != n:
            if Berry:
              Om_znk[ik,n] += -1.0*np.imag(pks[ik,jpol,n,m,0]*pks[ik,ipol,m,n,0]-pks[ik,ipol,n,m,0]*pks[ik,jpol,m,n,0])/((arrays['E_k'][ik,m,0] - arrays['E_k'][ik,n,0])**2 + deltab**2)
            if spin_Hall:
              Omj_znk[ik,n] += -2.0*np.imag(jks[ik,ipol,n,m,0]*pks[ik,jpol,m,n,0])/((arrays['E_k'][ik,m,0] - arrays['E_k'][ik,n,0])**2 + deltab**2)
      Om_zk[ik] = np.sum(Om_znk[ik,:]*(0.5 * (1 - np.sign(arrays['E_k'][ik,:bnd,0]))))  # T=0.0K
      if spin_Hall:
        Omj_zk[ik] = np.sum(Omj_znk[ik,:]*(0.5 * (1 - np.sign(arrays['E_k'][ik,:bnd,0]-mu))))  # T=0.0K

  indices = (LL[spol], LL[ipol], LL[jpol])
  lrng = (list(range(nkpi)) if rank==0 else None)

  pks = gather_full(pks, npool)
  velk = np.zeros((nkpi,3,bnd,nspin), dtype=float) if rank==0 else None
  if rank == 0:
    for n in range(bnd):
      velk[:,:,n,:] = np.real(pks[:,:,n,n,:])
  for l in range(3):
    fvk = 'velocity_'+str(l)
    data_controller.write_bands(fvk, (velk[:,l,:bnd,:] if rank==0 else None))
  pks = velk = None

  if Berry:
    Om_zk = gather_full(Om_zk, npool)
    fOm_zk = 'Omega_%s_%s%s.dat'%indices
    data_controller.write_file_row_col(fOm_zk, lrng, (-Om_zk[:,0] if rank==0 else None))
  Om_zk = fOm_zk = None

  if spin_Hall:
    Omj_zk = gather_full(Omj_zk, npool)
    fOmj_zk = 'Omegaj_%s_%s%s.dat'%indices
    data_controller.write_file_row_col(fOmj_zk, lrng, (Omj_zk[:,0] if rank==0 else None))
  Omj_zk = fOmj_zk = None


def band_loop_H ( HRaux, R, kq, nawf, nspin ):

  kdot = np.zeros((kq.shape[1],R.shape[0]),dtype=complex,order="C")
  kdot = np.tensordot(R,2.0j*np.pi*kq,axes=([1],[0]))
  np.exp(kdot,kdot)

  Haux = np.zeros((nawf,nawf,kq.shape[1],nspin),dtype=complex,order="C")

  for ispin in range(nspin):
    Haux[:,:,:,ispin] = np.tensordot(HRaux[:,:,:,ispin], kdot, axes=([2],[0]))

  kdot  = None
  Haux = np.transpose(Haux,(2,0,1,3))
  return Haux



#
# PAOFLOW
#
# Copyright 2016-2024 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

# This package is from: 
# M. Wimmer
# "Algorithm 923: Efficient Numerical Computation of the Pfaffian for
# Dense and Banded Skew-Symmetric Matrices".
# ACM Trans. Math. Software 38, 30 (2012).

"""A package for computing Pfaffians"""


import numpy as np
import scipy.linalg as la
import scipy.sparse as sp
import math, cmath

def householder_real(x):
    """(v, tau, alpha) = householder_real(x)

    Compute a Householder transformation such that
    (1-tau v v^T) x = alpha e_1
    where x and v a real vectors, tau is 0 or 2, and
    alpha a real number (e_1 is the first unit vector)
    """

    assert x.shape[0]>0

    sigma=np.dot(x[1:],x[1:])

    if sigma==0:
        return (np.zeros(x.shape[0]), 0, x[0])
    else:
        norm_x=math.sqrt(x[0]**2+sigma)

        v=x.copy();

        #depending on whether x[0] is positive or negatvie
        #choose the sign
        if x[0]<=0:
            v[0]-=norm_x
            alpha=+norm_x
        else:
            v[0]+=norm_x
            alpha=-norm_x

        v/=np.linalg.norm(v)

        return (v, 2, alpha)

def householder_complex(x):
    """(v, tau, alpha) = householder_real(x)

    Compute a Householder transformation such that
    (1-tau v v^T) x = alpha e_1
    where x and v a complex vectors, tau is 0 or 2, and
    alpha a complex number (e_1 is the first unit vector)
    """
    assert x.shape[0]>0

    sigma=np.dot(np.conj(x[1:]), x[1:])

    if sigma==0:
        return (np.zeros(x.shape[0]), 0, x[0])
    else:
        norm_x=cmath.sqrt(x[0].conjugate()*x[0]+sigma)

        v=x.copy();

        phase=cmath.exp(1j*math.atan2(x[0].imag, x[0].real))

        v[0]+=phase*norm_x

        v/=np.linalg.norm(v)

    return (v, 2, -phase*norm_x)

def skew_tridiagonalize(A, overwrite_a=False, calc_q=True):
    """ T, Q = skew_tridiagonalize(A, overwrite_a, calc_q=True)

    or

    T = skew_tridiagonalize(A, overwrite_a, calc_q=False)

    Bring a real or complex skew-symmetric matrix (A=-A^T) into
    tridiagonal form T (with zero diagonal) with a orthogonal
    (real case) or unitary (complex case) matrix U such that
    A = Q T Q^T
    (Note that Q^T and *not* Q^dagger also in the complex case)

    A is overwritten if overwrite_a=True (default: False), and
    Q only calculated if calc_q=True (default: True)
    """

    #Check if matrix is square
    assert A.shape[0] == A.shape[1] > 0
    #Check if it's skew-symmetric
    assert abs((A+A.T).max())<1e-14

    n = A.shape[0]
    A = np.asarray(A)  #the slice views work only properly for arrays

    #Check if we have a complex data type
    if np.issubdtype(A.dtype, np.complexfloating):
        householder = householder_complex
    elif not np.issubdtype(A.dtype, np.number):
        raise TypeError("pfaffian() can only work on numeric input")
    else:
        householder = householder_real

    if not overwrite_a:
        A = A.copy()

    if calc_q:
        Q = np.eye(A.shape[0], dtype=A.dtype)

    for i in range(A.shape[0]-2):
        #Find a Householder vector to eliminate the i-th column
        v, tau, alpha = householder(A[i+1:,i])
        A[i+1, i] = alpha
        A[i, i+1] = -alpha
        A[i+2:, i] = 0
        A[i, i+2:] = 0

        #Update the matrix block A(i+1:N,i+1:N)
        w = tau*np.dot(A[i+1:, i+1:], v.conj());
        A[i+1:,i+1:]+=np.outer(v,w)-np.outer(w,v)

        if calc_q:
            #Accumulate the individual Householder reflections
            #Accumulate them in the form P_1*P_2*..., which is
            # (..*P_2*P_1)^dagger
            y = tau*np.dot(Q[:, i+1:], v)
            Q[:, i+1:]-=np.outer(y,v.conj())

    if calc_q:
        return (np.asmatrix(A), np.asmatrix(Q))
    else:
        return np.asmatrix(A)

def skew_LTL(A, overwrite_a=False, calc_L=True, calc_P=True):
    """ T, L, P = skew_LTL(A, overwrite_a, calc_q=True)

    Bring a real or complex skew-symmetric matrix (A=-A^T) into
    tridiagonal form T (with zero diagonal) with a lower unit
    triangular matrix L such that
    P A P^T= L T L^T

    A is overwritten if overwrite_a=True (default: False),
    L and P only calculated if calc_L=True or calc_P=True,
    respectively (default: True).
    """

    #Check if matrix is square
    assert A.shape[0] == A.shape[1] > 0
    #Check if it's skew-symmetric
    assert abs((A+A.T).max())<1e-14

    n = A.shape[0]
    A = np.asarray(A)  #the slice views work only properly for arrays

    if not overwrite_a:
        A = A.copy()

    if calc_L:
        L = np.eye(n, dtype=A.dtype)

    if calc_P:
        Pv = np.arange(n)

    for k in range(n-2):
        #First, find the largest entry in A[k+1:,k] and
        #permute it to A[k+1,k]
        kp = k+1+np.abs(A[k+1:,k]).argmax()

        #Check if we need to pivot
        if kp != k+1:
            #interchange rows k+1 and kp
            temp = A[k+1,k:].copy()
            A[k+1,k:] = A[kp,k:]
            A[kp,k:] = temp

            #Then interchange columns k+1 and kp
            temp = A[k:,k+1].copy()
            A[k:,k+1] = A[k:,kp]
            A[k:,kp] = temp

            if calc_L:
                #permute L accordingly
                temp = L[k+1,1:k+1].copy()
                L[k+1,1:k+1] = L[kp,1:k+1]
                L[kp,1:k+1] = temp

            if calc_P:
                #accumulate the permutation matrix
                temp = Pv[k+1]
                Pv[k+1] = Pv[kp]
                Pv[kp] = temp

        #Now form the Gauss vector
        if A[k+1,k] != 0.0:
            tau = A[k+2:,k].copy()
            tau /= A[k+1,k]

            #clear eliminated row and column
            A[k+2:,k] = 0.0
            A[k,k+2:] = 0.0

            #Update the matrix block A(k+2:,k+2)
            A[k+2:,k+2:] += np.outer(tau, A[k+2:,k+1])
            A[k+2:,k+2:] -= np.outer(A[k+2:,k+1], tau)

            if calc_L:
                L[k+2:,k+1] = tau

    if calc_P:
        #form the permutation matrix as a sparse matrix
        P = sp.csr_matrix( (np.ones(n), (np.arange(n), Pv)) )

    if calc_L:
        if calc_P:
            return (np.asmatrix(A), np.asmatrix(L), P)
        else:
            return (np.asmatrix(A), np.asmatrix(L))
    else:
        if calc_P:
            return (np.asmatrix(A), P)
        else:
            return np.asmatrix(A)

def pfaffian(A, overwrite_a=False, method='P'):
    """ pfaffian(A, overwrite_a=False, method='P')

    Compute the Pfaffian of a real or complex skew-symmetric
    matrix A (A=-A^T). If overwrite_a=True, the matrix A
    is overwritten in the process. This function uses
    either the Parlett-Reid algorithm (method='P', default),
    or the Householder tridiagonalization (method='H')
    """
    #Check if matrix is square
    assert A.shape[0] == A.shape[1] > 0
    #Check if it's skew-symmetric
    assert abs((A+A.T).max())<1e-6
    #Check that the method variable is appropriately set
    assert method == 'P' or method == 'H'

    # Make sure the matrix is using floating point algebra
    if not np.issubdtype(A.dtype, np.inexact):
        A = 1.0 * A

    if method == 'P':
        return pfaffian_LTL(A, overwrite_a)
    else:
        return pfaffian_householder(A, overwrite_a)

def pfaffian_LTL(A, overwrite_a=False):
    """ pfaffian_LTL(A, overwrite_a=False)

    Compute the Pfaffian of a real or complex skew-symmetric
    matrix A (A=-A^T). If overwrite_a=True, the matrix A
    is overwritten in the process. This function uses
    the Parlett-Reid algorithm.
    """
    #Check if matrix is square
    assert A.shape[0] == A.shape[1] > 0
    #Check if it's skew-symmetric
    assert abs((A+A.T).max())<1e-14

    n = A.shape[0]
    A = np.asarray(A)  #the slice views work only properly for arrays

    #Quick return if possible
    if n%2==1:
        return 0

    if not overwrite_a:
        A = A.copy()

    pfaffian_val = 1.0

    for k in range(0, n-1, 2):
        #First, find the largest entry in A[k+1:,k] and
        #permute it to A[k+1,k]
        kp = k+1+np.abs(A[k+1:,k]).argmax()

        #Check if we need to pivot
        if kp != k+1:
            #interchange rows k+1 and kp
            temp = A[k+1,k:].copy()
            A[k+1,k:] = A[kp,k:]
            A[kp,k:] = temp

            #Then interchange columns k+1 and kp
            temp = A[k:,k+1].copy()
            A[k:,k+1] = A[k:,kp]
            A[k:,kp] = temp

            #every interchange corresponds to a "-" in det(P)
            pfaffian_val *= -1

        #Now form the Gauss vector
        if A[k+1,k] != 0.0:
            tau = A[k,k+2:].copy()
            tau /= A[k,k+1]

            pfaffian_val *= A[k,k+1]

            if k+2<n:
                #Update the matrix block A(k+2:,k+2)
                A[k+2:,k+2:] += np.outer(tau, A[k+2:,k+1])
                A[k+2:,k+2:] -= np.outer(A[k+2:,k+1], tau)
        else:
            #if we encounter a zero on the super/subdiagonal, the
            #Pfaffian is 0
            return 0.0

    return pfaffian_val


def pfaffian_householder(A, overwrite_a=False):
    """ pfaffian(A, overwrite_a=False)

    Compute the Pfaffian of a real or complex skew-symmetric
    matrix A (A=-A^T). If overwrite_a=True, the matrix A
    is overwritten in the process. This function uses the
    Householder tridiagonalization.

    Note that the function pfaffian_schur() can also be used in the
    real case. That function does not make use of the skew-symmetry
    and is only slightly slower than pfaffian_householder().
    """

    #Check if matrix is square
    assert A.shape[0] == A.shape[1] > 0
    #Check if it's skew-symmetric
    assert abs((A+A.T).max())<1e-14

    n = A.shape[0]

    #Quick return if possible
    if n%2==1:
        return 0

    #Check if we have a complex data type
    if np.issubdtype(A.dtype, np.complexfloating):
        householder=householder_complex
    elif not np.issubdtype(A.dtype, np.number):
        raise TypeError("pfaffian() can only work on numeric input")
    else:
        householder=householder_real

    A = np.asarray(A)  #the slice views work only properly for arrays

    if not overwrite_a:
        A = A.copy()

    pfaffian_val = 1.

    for i in range(A.shape[0]-2):
        #Find a Householder vector to eliminate the i-th column
        v, tau, alpha = householder(A[i+1:,i])
        A[i+1, i] = alpha
        A[i, i+1] = -alpha
        A[i+2:, i] = 0
        A[i, i+2:] = 0

        #Update the matrix block A(i+1:N,i+1:N)
        w = tau*np.dot(A[i+1:, i+1:], v.conj());
        A[i+1:,i+1:]+=np.outer(v,w)-np.outer(w,v)

        if tau!=0:
            pfaffian_val *= 1-tau
        if i%2==0:
            pfaffian_val *= -alpha

    pfaffian_val *= A[n-2,n-1]

    return pfaffian_val

def pfaffian_schur(A, overwrite_a=False):
    """Calculate Pfaffian of a real antisymmetric matrix using
    the Schur decomposition. (Hessenberg would in principle be faster,
    but scipy-0.8 messed up the performance for scipy.linalg.hessenberg()).

    This function does not make use of the skew-symmetry of the matrix A,
    but uses a LAPACK routine that is coded in FORTRAN and hence faster
    than python. As a consequence, pfaffian_schur is only slightly slower
    than pfaffian().
    """

    assert np.issubdtype(A.dtype, np.number) and not np.issubdtype(A.dtype, np.complexfloating)

    assert A.shape[0] == A.shape[1] > 0

    assert abs(A + A.T).max() < 1e-14

    #Quick return if possible
    if A.shape[0]%2 == 1:
        return 0

    (t, z) = la.schur(A, output='real', overwrite_a=overwrite_a)
    l = np.diag(t, 1)
    return np.prod(l[::2]) * la.det(z)



#
# PAOFLOW
#
# Copyright 2016-2022 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np
import scipy.sparse as sprs
import scipy.linalg as LA
from scipy.special import factorial as fac
from tempfile import NamedTemporaryFile
import re
from .communication import scatter_full, gather_full, gather_scatter
from scipy.spatial.distance import cdist
from mpi4py import MPI
from .zero_pad import zero_pad
import time

comm = MPI.COMM_WORLD
rank = comm.Get_rank()


def check(Hksp_s, si_per_k, new_k_ind, orig_k_ind, phase_shifts, U, a_index, inv_flag, equiv_atom, kp, symop, fg, isl,
          sym_TR):
  # compares saved full grid hamiltonain to one generated
  # from wedge... only for debugging purposes
  nawf = Hksp_s.shape[1]
  # load for testing
  Hksp_f = np.load("ONLY_FOR_TESTING_kham.npy")
  Hksp_f = np.reshape(Hksp_f, (nawf, nawf, Hksp_s.shape[0]))
  Hksp_f = np.transpose(Hksp_f, axes=(2, 0, 1))

  print(np.allclose(Hksp_f, Hksp_s, atol=1.e-4, rtol=1.e-4))

  bad_symop = np.ones(symop.shape[0], dtype=bool)
  good_symop = np.ones(symop.shape[0], dtype=bool)
  good = []
  bad = []
  print(a_index)
  st = 0
  fn = 8
  #    for j in range(20):
  for j in range(Hksp_s.shape[0]):
    isym = si_per_k[j]
    nki = new_k_ind[j]
    oki = orig_k_ind[j]

    HP = Hksp_f[nki]
    THP = Hksp_s[nki]

    U_k = get_U_k(kp[oki], phase_shifts[isym], a_index, U[isym])

    good.append(isym)

    if np.all(np.isclose(HP, THP, rtol=1.e-4, atol=1.e-4)):
      bad.append(isym)
    else:

      good_symop[isym] = False
      continue
      print(j, isym)
      print('old_k=', kp[oki])
      print('new_k=', fg[nki])
      print(equiv_atom[isym])
      print(symop[isym])
      print()
      print("inv:", inv_flag[isym])
      print()
      print("*" * 50)
      print("REAL " * 10)
      print("*" * 50)
      print()
      #           print(U_k[st:fn,st:fn])
      print()
      print("CORRECT")
      print(HP[st:fn, st:fn].real)
      print("NEW")
      print(THP[st:fn, st:fn].real)
      print("EQUAL?")
      print(np.isclose(HP[st:fn, st:fn].real, THP[st:fn, st:fn].real,
                       rtol=1.e-4, atol=1.e-4))

      print("*" * 50)
      print("IMAG " * 10)
      print("*" * 50)
      print("CORRECT")
      print(HP[st:fn, st:fn].imag)
      print("NEW")
      print(THP[st:fn, st:fn].imag)
      print("EQUAL?")
      print(np.isclose(HP[st:fn, st:fn].imag, THP[st:fn, st:fn].imag,
                       rtol=1.e-4, atol=1.e-4))

      print("*" * 100)
      print("*" * 100)
      print("*" * 100)
  #            raise SystemExit
  print(len(good) - kp.shape[0], len(bad) - kp.shape[0])

  print()
  for i in range(symop.shape[0]):
    if i in si_per_k:
      if not good_symop[i]:
        print("%2d" % (i + 1), "BAD ", isl[i], sym_TR[i])
      else:
        print("%2d" % (i + 1), "GOOD", isl[i], sym_TR[i])


############################################################################################
############################################################################################
############################################################################################

def LPF(aux, nk1, nk2, nk3, nfft1, nfft2, nfft3):
  '''
  Pad frequency domain with zeroes, such that any relationship between
      aux[k] and aux[N-k] is preserved.

  Arguments:
      aux (ndarray): unpadded frequency domain data
      nk1 (int): current size of aux along axis 0
      nk2 (int): current size of aux along axis 1
      nk3 (int): current size of aux along axis 2
      nfft1 (int): number of zeroes to pad axis 0 by
      nfft1 (int): number of zeroes to pad axis 1 by
      nfft1 (int): number of zeroes to pad axis 2 by

  Returns:
      auxp3 (ndarray): padded frequency domain data
  '''

  # post-padding dimensions
  nk1p = nfft1 + nk1
  nk2p = nfft2 + nk2
  nk3p = nfft3 + nk3
  # halfway points
  sk1 = int((nk1 + 1) / 2)
  sk2 = int((nk2 + 1) / 2)
  sk3 = int((nk3 + 1) / 2)
  # parities (even <-> p==1)
  p1 = (nk1 & 1) ^ 1
  p2 = (nk2 & 1) ^ 1
  p3 = (nk3 & 1) ^ 1

  # accomodate nfft==0
  if nfft1 == 0:  p1 = 0
  if nfft2 == 0:  p2 = 0
  if nfft3 == 0:  p3 = 0

  # first dimension
  auxp1 = np.zeros((nk1, nk2, nk3p), dtype=complex)
  auxp1[:, :, :sk3 + p3] = aux[:, :, :sk3 + p3]
  auxp1[:, :, nfft3 + sk3:] = aux[:, :, sk3:]
  # second dimension
  auxp2 = np.zeros((nk1, nk2p, nk3p), dtype=complex)
  auxp2[:, :sk2 + p2, :] = auxp1[:, :sk2 + p2, :]
  auxp2[:, nfft2 + sk2:, :] = auxp1[:, sk2:, :]
  # third dimension
  auxp3 = np.zeros((nk1p, nk2p, nk3p), dtype=complex)
  auxp3[:sk1 + p1, :, :] = auxp2[:sk1 + p1, :, :]
  auxp3[nfft1 + sk1:, :, :] = auxp2[sk1:, :, :]

  if rank == 0:
    print(sk1)
  # halve Nyquist axes
  if p1:
    auxp3[sk1 - 1, :, :] = 0.0
    auxp3[-sk1 + 1, :, :] = 0.0
    auxp3[sk1, :, :] = 0.0
  if p2:
    auxp3[:, sk2 - 1, :] = 0.0
    auxp3[:, -sk2 + 1, :] = 0.0
    auxp3[:, -sk2, :] = 0.0
  if p3:
    auxp3[:, :, sk3 - 1] = 0.0
    auxp3[:, :, -sk3 + 1] = 0.0
    auxp3[:, :, -sk3] = 0.0
  return (auxp3)


############################################################################################
############################################################################################
############################################################################################

def down_samp(aux, nk1, nk2, nk3, nfft1, nfft2, nfft3):
  '''
  Pad frequency domain with zeroes, such that any relationship between
      aux[k] and aux[N-k] is preserved.

  Arguments:
      aux (ndarray): unpadded frequency domain data
      nk1 (int): current size of aux along axis 0
      nk2 (int): current size of aux along axis 1
      nk3 (int): current size of aux along axis 2
      nfft1 (int): number of zeroes to pad axis 0 by
      nfft1 (int): number of zeroes to pad axis 1 by
      nfft1 (int): number of zeroes to pad axis 2 by

  Returns:
      auxp3 (ndarray): padded frequency domain data
  '''

  nk1 += nfft1
  nk2 += nfft2
  nk3 += nfft3

  nfft1 *= -1
  nfft2 *= -1
  nfft3 *= -1

  # post-padding dimensions
  nk1p = nfft1 + nk1
  nk2p = nfft2 + nk2
  nk3p = nfft3 + nk3
  # halfway points
  sk1 = int((nk1 + 1) / 2)
  sk2 = int((nk2 + 1) / 2)
  sk3 = int((nk3 + 1) / 2)
  # parities (even <-> p==1)
  p1 = (nk1 & 1) ^ 1
  p2 = (nk2 & 1) ^ 1
  p3 = (nk3 & 1) ^ 1

  # accomodate nfft==0
  if nfft1 == 0:  p1 = 0
  if nfft2 == 0:  p2 = 0
  if nfft3 == 0:  p3 = 0

  # first dimension
  auxp1 = np.zeros((nk1p, nk2p, nk3), dtype=complex)
  auxp1[:, :, :sk3 + p3] = aux[:, :, :sk3 + p3]
  auxp1[:, :, sk3:] = aux[:, :, nfft3 + sk3:]
  # second dimension
  auxp2 = np.zeros((nk1p, nk2, nk3), dtype=complex)
  auxp2[:, :sk2 + p2, :] = auxp1[:, :sk2 + p2, :]
  auxp2[:, sk2:, :] = auxp1[:, nfft2 + sk2:, :]
  # third dimension
  auxp3 = np.zeros((nk1, nk2, nk3), dtype=complex)
  auxp3[:sk1 + p1, :, :] = auxp2[:sk1 + p1, :, :]
  auxp3[sk1:, :, :] = auxp2[nfft1 + sk1:, :, :]

  # double Nyquist axes
  if p1:
    auxp3[sk1, :, :] *= 2
  if p2:
    auxp3[:, sk2, :] *= 2
  if p3:
    auxp3[:, :, sk3] *= 2

  return auxp3


############################################################################################
############################################################################################
############################################################################################

def map_equiv_atoms(a_index, map_ind):
  # generate unitary tranformation for swapping
  # between atoms equivalent by symmetry op

  nawf = a_index.shape[0]
  U_wyc = np.zeros((map_ind.shape[0], nawf, nawf), dtype=complex)

  for isym in range(map_ind.shape[0]):
    map_U = np.zeros((nawf, nawf), dtype=complex)
    for i in range(map_ind[isym].shape[0]):
      fi = np.where(a_index == i)[0]
      si = np.where(a_index == map_ind[isym][i])[0]

      for j in range(si.shape[0]):
        map_U[fi[j], si[j]] = 1

    U_wyc[isym] = map_U

  return U_wyc


############################################################################################
############################################################################################
############################################################################################

def get_U_TR(jchia):
  # operator that swaps J for -J

  J = 0.5
  TR_J05 = np.fliplr(np.eye(int(2 * J + 1)) * ((-1) ** (J - np.arange(-J, J + 1))))
  J = 1.5
  TR_J15 = np.fliplr(np.eye(int(2 * J + 1)) * ((-1) ** (J - np.arange(-J, J + 1))))
  J = 2.5
  TR_J25 = np.fliplr(np.eye(int(2 * J + 1)) * ((-1) ** (J - np.arange(-J, J + 1))))
  J = 3.5
  TR_J35 = np.fliplr(np.eye(int(2 * J + 1)) * ((-1) ** (J - np.arange(-J, J + 1))))

  TR = [TR_J05, TR_J15, TR_J25, TR_J35]

  nawf = int(np.sum(2 * jchia + 1) * 2)

  U_TR = np.zeros((nawf, nawf), dtype=complex)

  blocks = []

  flr_jchia = np.array([int(np.floor(a)) for a in jchia])

  # block diagonal transformation matrix of T @ H(k) @ T^-1
  for s in flr_jchia:
    blocks.append(TR[s])
  U_TR = LA.block_diag(*blocks)

  return U_TR


############################################################################################
############################################################################################
############################################################################################

def add_U_TR(U, sym_TR, jchia):
  # applies time inversion to symops as needed
  U_TR = get_U_TR(jchia)

  for isym in range(U.shape[0]):
    if sym_TR[isym]:
      U[isym] = U_TR @ U[isym]

  return U


############################################################################################
############################################################################################
############################################################################################


def get_trans():
  # get transformation from angular momentum number to
  # chemistry orbital form for the angular momentum
  rsh = np.sqrt(0.5)
  ish = 1.0j * np.sqrt(0.5)

  trans_l0 = np.array([[1.0]], dtype=complex)

  # to convert from m=[-1,0,1] to px,py,pz for l=1
  trans_l1 = np.array([[0, 1, 0],
                       [-1, 0, 1],
                       [1, 0, 1], ], dtype=complex)

  trans_l1[1] *= rsh
  trans_l1[2] *= ish

  # to convert from m=[-2,-1,0,1,2] to dx2,dzx,dzy,dx2-y2,dxy for l=2
  trans_l2 = np.array([[0, 0, 1, 0, 0],
                       [0, -1, 0, 1, 0],
                       [0, 1, 0, 1, 0],
                       [1, 0, 0, 0, 1],
                       [-1, 0, 0, 0, 1], ], dtype=complex)

  trans_l2[1] *= rsh
  trans_l2[2] *= ish
  trans_l2[3] *= rsh
  trans_l2[4] *= ish

  # to convert from m=[-3,-2,-1,0,1,2,3] to real combinations for l=3
  trans_l3 = np.array([[0, 0, 0, 1, 0, 0, 0],
                       [0, 0, -1, 0, 1, 0, 0],
                       [0, 0, 1, 0, 1, 0, 0],
                       [0, 1, 0, 0, 0, 1, 0],
                       [0, -1, 0, 0, 0, 1, 0],
                       [-1, 0, 0, 0, 0, 0, 1],
                       [1, 0, 0, 0, 0, 0, 1], ], dtype=complex)

  trans_l3[1] *= rsh
  trans_l3[2] *= ish
  trans_l3[3] *= rsh
  trans_l3[4] *= ish
  trans_l3[5] *= rsh
  trans_l3[6] *= ish

  return trans_l0, trans_l1, trans_l2, trans_l3


############################################################################################
############################################################################################
############################################################################################


def d_mat_l(AL, BE, GA, l):
  # gets wigner_d matrix for a given alpha,beta,gamma
  # that rotates basis from m' to m

  # list of m and m'
  ms = np.arange(-l, l + 1)
  mps = ms

  d_mat = np.zeros((ms.shape[0], ms.shape[0]), dtype=complex)

  # find wigner_d matrix elements for each m and m'
  for m_i in range(len(ms)):
    for mp_i in range(len(mps)):
      m = ms[m_i]
      mp = mps[mp_i]
      w_max = int(l + mp + 1)
      dmm = 0.0

      out_sum = np.sqrt(fac(int(l + m)) * fac(int(l - m)) * fac(int(l + mp)) * fac(int(l - mp)))

      # loop over w for summation
      for w in range(0, w_max):
        # factorials in denominator must be positive
        df1 = int(l + mp - w)
        df2 = int(l - m - w)
        df3 = int(w)
        df4 = int(w + m - mp)
        if df1 < 0 or df2 < 0 or df3 < 0 or df4 < 0: continue

        # for each w in summation over w
        tmp = (-1) ** w
        tmp /= fac(df1) * fac(df2) * fac(df3) * fac(df4)
        tmp *= np.cos(BE / 2.0) ** (2 * l - m + mp - 2 * w)
        tmp *= np.sin(BE / 2.0) ** (2 * w - mp + m)

        dmm += tmp

      d_mat[mp_i, m_i] = dmm * np.exp(-1.0j * AL * mp) * np.exp(-1.0j * GA * m) * out_sum

  return d_mat


############################################################################################
############################################################################################
############################################################################################

def get_wigner(symop):
  # gets wigner_d matrix associated with each symop
  wigner_l0 = np.zeros((symop.shape[0], 1, 1), dtype=complex)
  wigner_l1 = np.zeros((symop.shape[0], 3, 3), dtype=complex)
  wigner_l2 = np.zeros((symop.shape[0], 5, 5), dtype=complex)
  wigner_l3 = np.zeros((symop.shape[0], 7, 7), dtype=complex)
  # inversion flag
  inv_flag = np.zeros((symop.shape[0]), dtype=bool)

  for i in range(symop.shape[0]):
    # get euler angles alpha,beta,gamma from the symop
    AL, BE, GA = mat2eul(symop[i])
    AL, BE, GA = np.deg2rad(np.around(np.rad2deg([AL, BE, GA]), decimals=0))

    # check if there is an inversion in the symop
    if not np.all(np.isclose(eul2mat(AL, BE, GA), symop[i])):
      inv_flag[i] = True
      AL, BE, GA = mat2eul(-symop[i])
      AL, BE, GA = np.deg2rad(np.around(np.rad2deg([AL, BE, GA]), decimals=0))
      if not np.all(np.isclose(eul2mat(AL, BE, GA), -symop[i])):
        print("ERROR IN MAT2EUL!")
        print(i + 1)
        print(symop[i])
        raise SystemExit

    # wigner_d matrix for l=0
    wigner_l0[i] = d_mat_l(AL, BE, GA, 0)
    # wigner_d matrix for l=1
    wigner_l1[i] = d_mat_l(AL, BE, GA, 1)
    # wigner_d matrix for l=2
    wigner_l2[i] = d_mat_l(AL, BE, GA, 2)
    # wigner_d matrix for l=3
    wigner_l3[i] = d_mat_l(AL, BE, GA, 3)

  return [wigner_l0, wigner_l1, wigner_l2, wigner_l3, inv_flag], inv_flag


############################################################################################
############################################################################################
############################################################################################

def get_wigner_so(symop):
  # gets wigner_d matrix associated with each symop
  wigner_j05 = np.zeros((symop.shape[0], 2, 2), dtype=complex)
  wigner_j15 = np.zeros((symop.shape[0], 4, 4), dtype=complex)
  wigner_j25 = np.zeros((symop.shape[0], 6, 6), dtype=complex)
  wigner_j35 = np.zeros((symop.shape[0], 8, 8), dtype=complex)
  # inversion flag
  inv_flag = np.zeros((symop.shape[0]), dtype=bool)

  for i in range(symop.shape[0]):
    # get euler angles alpha,beta,gamma from the symop
    AL, BE, GA = mat2eul(symop[i])
    AL, BE, GA = np.deg2rad(np.around(np.rad2deg([AL, BE, GA]), decimals=0))
    # check if there is an inversion in the symop
    if not np.all(np.isclose(correct_roundoff(eul2mat(AL, BE, GA)),
                             symop[i], atol=1.e-3, rtol=1.e-2)):
      inv_flag[i] = True

      AL, BE, GA = mat2eul(-symop[i])
      AL, BE, GA = np.deg2rad(np.around(np.rad2deg([AL, BE, GA]), decimals=0))
      if not np.all(np.isclose(correct_roundoff(eul2mat(AL, BE, GA)),
                               -symop[i], atol=1.e-3, rtol=1.e-2)):
        print("ERROR IN MAT2EUL!")
        print(i + 1)
        print('RESULT')
        print(-eul2mat(AL, BE, GA))
        print('CORRECT')
        print(symop[i])
        raise SystemExit

    # wigner_d for l=0
    wigner_j05[i] = d_mat_l(AL, BE, GA, 0.5)
    # wigner_d for l=1
    wigner_j15[i] = d_mat_l(AL, BE, GA, 1.5)
    # wigner_d for l=2
    wigner_j25[i] = d_mat_l(AL, BE, GA, 2.5)
    # wigner_d for l=3
    wigner_j35[i] = d_mat_l(AL, BE, GA, 3.5)

  return [wigner_j05, wigner_j15, wigner_j25, wigner_j35, inv_flag], inv_flag


############################################################################################
############################################################################################
############################################################################################

def convert_wigner_d(wigner):
  # get transformation from angular momentum number to
  # chemistry orbital form for the angular momentum

  trans_l0, trans_l1, trans_l2, trans_l3 = get_trans()

  c_wigner_l0 = np.zeros_like(wigner[0])
  c_wigner_l1 = np.zeros_like(wigner[1])
  c_wigner_l2 = np.zeros_like(wigner[2])
  c_wigner_l3 = np.zeros_like(wigner[3])

  for i in range(wigner[0].shape[0]):
    c_wigner_l0[i] = trans_l0 @ wigner[0][i] @ np.conj(trans_l0.T)
    c_wigner_l1[i] = trans_l1 @ wigner[1][i] @ np.conj(trans_l1.T)
    c_wigner_l2[i] = trans_l2 @ wigner[2][i] @ np.conj(trans_l2.T)
    c_wigner_l3[i] = trans_l3 @ wigner[3][i] @ np.conj(trans_l3.T)

  return [c_wigner_l0, c_wigner_l1, c_wigner_l2, c_wigner_l3]


############################################################################################
############################################################################################
############################################################################################

def eul2mat(ALPHA, BETA, GAMMA):
  # generates rotation matrix from euler angles
  C_1 = np.cos(ALPHA)
  C_2 = np.cos(BETA)
  C_3 = np.cos(GAMMA)
  S_1 = np.sin(ALPHA)
  S_2 = np.sin(BETA)
  S_3 = np.sin(GAMMA)

  euler = np.array([[C_1 * C_2 * C_3 - S_1 * S_3, -C_1 * C_2 * S_3 - C_3 * S_1, C_1 * S_2, ],
                    [C_2 * C_3 * S_1 + C_1 * S_3, -C_2 * S_1 * S_3 + C_1 * C_3, S_1 * S_2, ],
                    [-C_3 * S_2, S_2 * S_3, C_2, ]])

  return euler


############################################################################################
############################################################################################
############################################################################################

def mat2eul(R):
  # finds euler angles from rotation matrix in ZYZ convention
  if R[2, 2] < 1.0:
    if R[2, 2] > -1.0:
      ALPHA = np.arctan2(R[1, 2], R[0, 2])
      BETA = np.arccos(R[2, 2])
      GAMMA = np.arctan2(R[2, 1], -R[2, 0])
    else:
      ALPHA = -np.arctan2(R[1, 0], R[1, 1])
      BETA = np.pi
      GAMMA = 0.0
  else:
    ALPHA = np.arctan2(R[1, 0], R[1, 1])
    BETA = 0.0
    GAMMA = 0.0

  return np.array([ALPHA, BETA, GAMMA])


############################################################################################
############################################################################################
############################################################################################

def find_equiv_k(kp, symop, full_grid, sym_TR, check=True, include_self=False):
  # find indices and symops that generate full grid H from wedge H
  orig_k_ind = []
  new_k_ind = []
  si_per_k = []
  counter = 0
  kp_track = []
  kp = correct_roundoff(kp)

  full_grid_mask = np.copy(full_grid)

  for isym in range(symop.shape[0]):
    # transform k -> k' with the sym op
    if sym_TR[isym]:
      newk = ((((-symop[isym] @ (kp.T % 1.0)) % 1.0) + 0.5) % 1.0) - 0.5
    else:
      newk = ((((symop[isym] @ (kp.T % 1.0)) % 1.0) + 0.5) % 1.0) - 0.5
    newk = correct_roundoff(newk)
    newk[np.where(np.isclose(newk, 0.5))] = -0.5
    newk[np.where(np.isclose(newk, -1.0))] = 0.0
    newk[np.where(np.isclose(newk, 1.0))] = 0.0

    # find index in the full grid where this k -> k' with this sym op
    nw = np.where(np.isclose(cdist(newk.T, full_grid_mask), 0.0, atol=1.e-6, rtol=1.e-6, ))

    if not include_self:
      full_grid_mask[nw[1]] = np.ma.masked

    new_k_ind.extend(nw[1].tolist())
    si_per_k.extend([isym] * nw[1].shape[0])
    orig_k_ind.extend(nw[0].tolist())

  new_k_ind = np.array(new_k_ind)
  orig_k_ind = np.array(orig_k_ind)
  si_per_k = np.array(si_per_k)

  if not include_self:
    inds = np.unique(new_k_ind, return_index=True)

    new_k_ind = new_k_ind[inds[1]]
    orig_k_ind = orig_k_ind[inds[1]]
    si_per_k = si_per_k[inds[1]]
    counter = inds[1].shape[0]

  # check to make sure we have all the k points accounted for
  if counter != full_grid.shape[0] and check:

    print('NOT ALL KPOINTS ACCOUNTED FOR')
    print(counter, full_grid.shape[0])
    print('missing k:')
    print(full_grid[np.setxor1d(new_k_ind, np.arange(full_grid.shape[0]))])
    raise SystemExit

  else:
    pass

  return new_k_ind, orig_k_ind, si_per_k


############################################################################################
############################################################################################
############################################################################################


def build_U_matrix(wigner, shells):
  # builds U from blocks

  nawf = int(np.sum(2 * shells + 1))

  U = np.zeros((wigner[0].shape[0], nawf, nawf), dtype=complex)
  # for the so case...
  flr_shells = np.array([int(np.floor(a)) for a in shells])

  for i in range(wigner[0].shape[0]):
    blocks = []
    # block diagonal transformation matrix of H(k)->H(k')
    for s in flr_shells:
      blocks.append(wigner[s][i])
    U[i] = LA.block_diag(*blocks)
  return U


############################################################################################
############################################################################################
############################################################################################

def get_phase_shifts(atom_pos, symop, equiv_atom):
  # calculate phase shifts for U
  phase_shift = np.zeros((symop.shape[0], atom_pos.shape[0], 3), dtype=float)
  for isym in range(symop.shape[0]):
    for p in range(atom_pos.shape[0]):
      p1 = equiv_atom[isym, p]
      phase_shift[isym, p1] = (symop[isym].T @ atom_pos[p]) - atom_pos[p1]

  return correct_roundoff(np.around(phase_shift, decimals=2))


############################################################################################
############################################################################################
############################################################################################

def get_U_k(k, shift, a_index, U):
  # add phase shift to U
  U_k = U * np.exp(2.0j * np.pi * (shift[a_index] @ k))
  return U_k


############################################################################################
############################################################################################
############################################################################################

def correct_roundoff(arr, incl_hex=False, atol=1.e-8):
  # correct for round off
  arr[np.where(np.isclose(arr, 0.0, atol=atol))] = 0.0
  arr[np.where(np.isclose(arr, 1.0, atol=atol))] = 1.0
  arr[np.where(np.isclose(arr, -1.0, atol=atol))] = -1.0

  if incl_hex:
    sq3o2 = np.sqrt(3) / 2.0
    arr[np.where(np.isclose(arr, sq3o2, atol=atol))] = sq3o2
    arr[np.where(np.isclose(arr, -sq3o2, atol=atol))] = -sq3o2
    arr[np.where(np.isclose(arr, 0.5, atol=atol))] = 0.5
    arr[np.where(np.isclose(arr, -0.5, atol=atol))] = -0.5

  return arr


############################################################################################
############################################################################################
############################################################################################

def get_full_grid(nk1, nk2, nk3, o1, o2, o3):
  # generates full k grid in crystal fractional coords
  nktot = nk1 * nk2 * nk3
  b_vectors = np.eye(3)
  Kint = np.zeros((nktot, 3), dtype=float)
  offset = np.array([o1, o2, o3]) / (2 * np.array([nk1, nk2, nk3]))

  for i in range(nk1):
    for j in range(nk2):
      for k in range(nk3):
        n = k + j * nk3 + i * nk2 * nk3
        Rx = float(i) / float(nk1)
        Ry = float(j) / float(nk2)
        Rz = float(k) / float(nk3)
        if Rx >= 0.5: Rx = Rx - 1.0
        if Ry >= 0.5: Ry = Ry - 1.0
        if Rz >= 0.5: Rz = Rz - 1.0
        Rx -= int(Rx)
        Ry -= int(Ry)
        Rz -= int(Rz)

        Kint[n] = Rx * b_vectors[0, :] + Ry * b_vectors[1, :] + Rz * b_vectors[2, :] + offset

  return Kint


############################################################################################
############################################################################################
############################################################################################

def get_inv_op(shells):
  # returns the inversion operator
  orb_index = np.hstack([[l] * (2 * l + 1) for l in shells])
  sign_inv = (-np.ones(orb_index.shape[0])) ** orb_index

  return np.outer(sign_inv, sign_inv)


############################################################################################
############################################################################################
############################################################################################

def enforce_t_rev(Hksp_s, nk1, nk2, nk3, spin_orb, U_inv, jchia):
  # enforce time reversal symmetry on H(k)
  nawf = Hksp_s.shape[1]

  Hksp_s = np.reshape(Hksp_s, (nk1, nk2, nk3, nawf, nawf), order="C")

  if spin_orb:
    U_TR = get_U_TR(jchia)

  for i in range(int(nk1 / 2) + 1):
    for j in range(int(nk2 / 2) + 1):
      for k in range(int(nk3 / 2) + 1):
        iv = (nk1 - i) % nk1
        jv = (nk2 - j) % nk2
        kv = (nk3 - k) % nk3
        if not spin_orb:
          temp1 = np.conj(Hksp_s[i, j, k])
          temp2 = np.conj(Hksp_s[iv, jv, kv])
          Hksp_s[iv, jv, kv] = (Hksp_s[iv, jv, kv] + temp1) / 2.0
          Hksp_s[i, j, k] = (Hksp_s[i, j, k] + temp2) / 2.0
        else:
          temp1 = np.conj(U_inv * (U_TR @ Hksp_s[i, j, k] @ np.conj(U_TR.T)))
          #                    temp2=np.conj(U_inv*(U_TR @ Hksp_s[iv,jv,kv] @ np.conj(U_TR.T)))
          #                    Hksp_s[iv,jv,kv] = (Hksp_s[iv,jv,kv] + temp1)/2.0
          #                    Hksp_s[i,j,k]    = (Hksp_s[i,j,k]    + temp2)/2.0
          Hksp_s[iv, jv, kv] = temp1

  Hksp_s = np.reshape(Hksp_s, (nk1 * nk2 * nk3, nawf, nawf), order="C")

  return Hksp_s


############################################################################################
############################################################################################
############################################################################################

def apply_t_rev(Hksp, kp, spin_orb, U_inv, jchia):
  # apply time reversal operator to get H(-k) from H(k)

  # gamma only case
  if Hksp.shape[0] == 1:
    return Hksp, kp
  new_kp_list = []
  new_Hk_list = []

  if spin_orb:
    U_TR = get_U_TR(jchia)

  for i in range(Hksp.shape[0]):
    new_kp = -kp[i]
    if not np.any(np.all(np.isclose(new_kp, kp), axis=1)):
      new_kp_list.append(new_kp)
      if not spin_orb:
        new_Hk_list.append(np.conj(Hksp[i]))
      else:
        #                print(U_inv.shape,U_TR.shape,Hksp.shape)
        new_Hk_list.append(np.conj(U_inv * (U_TR @ Hksp[i] @ np.conj(U_TR.T))))

  if len(new_kp_list) == 0:
    return Hksp, kp

  kp = np.vstack([kp, np.array(new_kp_list)])
  Hksp = np.vstack([Hksp, np.array(new_Hk_list)])

  return Hksp, kp


############################################################################################
############################################################################################
############################################################################################

def enforce_hermaticity(Hksp):
  # enforce H(k) to be hermitian (it should be already)
  for k in range(Hksp.shape[0]):
    Hksp[k] = (Hksp[k] + np.conj(Hksp[k].T)) / 2.0

  return Hksp


############################################################################################
############################################################################################
############################################################################################

def add_U_wyc(U, U_wyc):
  # add operator that shifts orbital from one equivalent site to another
  for isym in range(U.shape[0]):
    U[isym] = U_wyc[isym] @ U[isym]
    U[isym] = correct_roundoff(U[isym], incl_hex=True)

  return U


############################################################################################
############################################################################################
############################################################################################

def wedge_to_grid(Hksp, U, a_index, phase_shifts, kp, new_k_ind, orig_k_ind, si_per_k, inv_flag, U_inv, sym_TR,
                  spin_orb, npool):
  # generates full grid from k points in IBZ
  nawf = Hksp.shape[1]

  fgm = scatter_full(np.arange(si_per_k.shape[0], dtype=int), npool)
  new_k_ind = scatter_full(new_k_ind, npool)
  orig_k_ind = scatter_full(orig_k_ind, npool)
  si_per_k = scatter_full(si_per_k, npool)

  Hksp_s = np.zeros((new_k_ind.shape[0], nawf, nawf), dtype=complex)

  for j in range(new_k_ind.shape[0]):
    isym = si_per_k[j]
    oki = orig_k_ind[j]
    nki = np.where(new_k_ind[j] == fgm)[0]

    # if symop is identity
    if isym == 0:
      Hksp_s[nki] = Hksp[oki]
      continue

    # other cases
    H = Hksp[oki]

    # get k dependent U
    U_k = get_U_k(kp[oki], phase_shifts[isym], a_index, U[isym])

    # transformated H(k)
    THP = U_k @ H @ np.conj(U_k.T)

    # apply inversion operator if needed
    if inv_flag[isym]:
      THP *= U_inv

    # time inversion is anti-unitary
    # old version
    # if sym_TR[isym]:
    #    THP *= U_inv
    #    THP = np.conj(THP)
    if sym_TR[isym]:
      if spin_orb:
        THP *= U_inv
      THP = np.conj(THP)

    Hksp_s[nki] = THP

  # make sure of hermiticity of each H(k)
  Hksp_s = enforce_hermaticity(Hksp_s)

  Hksp = None
  Hksp_s = gather_full(Hksp_s, npool)

  return Hksp_s


############################################################################################
############################################################################################
############################################################################################

def open_grid(Hksp, full_grid, kp, symop, symop_cart, atom_pos, shells, a_index, equiv_atom,
              nk1, nk2, nk3, o1, o2, o3, spin_orb, sym_TR, jchia, mag_calc, symm_grid, thresh, max_iter, verbose, npool):
  # calculates full H(k) grid from wedge

  nawf = Hksp.shape[1]
  # get inversion operator
  U_inv = get_inv_op(shells)

  # apply time reversal symmetry H(k) = H(-k)* where appropriate
  if not (spin_orb and mag_calc):
    Hksp, kp = apply_t_rev(Hksp, kp, spin_orb, U_inv, jchia)

  # get array with wigner_d rotation matrix for each symop
  # for each of the orbital angular momentum l=[0,1,2,3]
  if spin_orb:
    wigner, inv_flag = get_wigner_so(symop_cart)
  else:
    wigner, inv_flag = get_wigner(symop_cart)
    # convert the wigner_d into chemistry form for each symop
    wigner = convert_wigner_d(wigner)

  # get phase shifts from rotation symop
  phase_shifts = get_phase_shifts(atom_pos, symop, equiv_atom)

  # build U and U_inv from blocks
  if spin_orb:
    U = build_U_matrix(wigner, jchia)
  else:
    U = build_U_matrix(wigner, shells)

  # if any symop involve time inversion add the TR to the symop
  if np.any(sym_TR) and spin_orb:
    U = add_U_TR(U, sym_TR, jchia)

  # adds transformation to U that maps orbitals from
  # atom A to equivalent atom B atoms after symop
  U_wyc = map_equiv_atoms(a_index, equiv_atom)

  # combine U_wyc and U
  U = add_U_wyc(U, U_wyc)

  # get index of k in wedge, index in full grid,
  # and index of symop that transforms k to k'
  new_k_ind, orig_k_ind, si_per_k = find_equiv_k(kp, symop, full_grid, sym_TR, check=True)

  # transform H(k) -> H(k')
  Hksp = wedge_to_grid(Hksp, U, a_index, phase_shifts, kp,
                       new_k_ind, orig_k_ind, si_per_k, inv_flag, U_inv, sym_TR, spin_orb, npool)

  if rank == 0:
    # enforce time reversion where appropriate
    if not (spin_orb and mag_calc):
      Hksp = enforce_t_rev(Hksp, nk1, nk2, nk3, spin_orb, U_inv, jchia)

  else:
    Hksp = np.zeros((full_grid.shape[0], nawf, nawf), dtype=complex)

  comm.Bcast(Hksp)

  if symm_grid:

    symop_inv = np.zeros_like(symop)
    for i in range(symop.shape[0]):
      symop_inv[i] = LA.inv(symop[i])

    nkl = []
    partial_grid = scatter_full(full_grid, npool)
    for i in range(partial_grid.shape[0]):
      nkl.append(find_equiv_k(partial_grid[i][None], symop_inv, full_grid, sym_TR, check=False, include_self=True))
    nkl_no_interp = np.array(nkl)
    Hksp, tmax = symmetrize_grid(Hksp, U, a_index, phase_shifts, inv_flag, U_inv, sym_TR,
                                 full_grid, jchia, spin_orb, mag_calc, nk1, nk2, nk3,
                                 nkl_no_interp, partial_grid, npool)
    upscale1 = int(0.25 * nk1)
    upscale2 = int(0.25 * nk2)
    upscale3 = int(0.25 * nk3)

    nfft1 = nk1 + upscale1
    nfft2 = nk2 + upscale2
    nfft3 = nk3 + upscale3

    full_grid_interp = get_full_grid(nfft1, nfft2, nfft3, o1, o2, o3)
    nkl = []
    partial_grid_interp = scatter_full(full_grid_interp, npool)
    for i in range(partial_grid_interp.shape[0]):
      nkl.append(
        find_equiv_k(partial_grid_interp[i][None], symop_inv, full_grid_interp, sym_TR, check=False, include_self=True))
    nkl_interp = np.array(nkl)
    # max difference bewtween H(k) and H(k*)
    tmax = 999999

    for i in range(int(max_iter * 2)):
      st = time.time()
      add1 = upscale1 * ((-1) ** i)
      add2 = upscale2 * ((-1) ** i)
      add3 = upscale3 * ((-1) ** i)

      nfft1 = nk1 + add1
      nfft2 = nk2 + add2
      nfft3 = nk3 + add3

      if rank == 0:
        Hksp = np.reshape(Hksp, (nk1 * nk2 * nk3, nawf * nawf))
        Hksp = np.ascontiguousarray(Hksp.T)

      Hksp = scatter_full(Hksp, npool)

      Hksp = np.reshape(Hksp, (Hksp.shape[0], nk1, nk2, nk3))
      HRs = np.fft.ifftn(Hksp, axes=(1, 2, 3))

      switch = True
      if switch == True:

        Hksp = None
        Hksp = np.zeros((HRs.shape[0], nfft1, nfft2, nfft3), dtype=complex)

        for m in range(Hksp.shape[0]):
          Hksp[m, :, :, :] = np.fft.fftn(zero_pad(HRs[m, :, :, :], nk1, nk2, nk3, add1, add2, add3))
        #                     if not i%2:
        #                         Hksp[m,:,:,:]=np.fft.fftn(zero_pad(HRs[m,:,:,:],nk1,nk2,nk3,add1,add2,add3))
        #                     else:
        # #                        Hksp[m,:,:,:]=np.fft.fftn(LPF(HRs[m,:,:,:],nk1,nk2,nk3,add1,add2,add3))
        #                         Hksp[m,:,:,:]=np.fft.fftn(down_samp(HRs[m,:,:,:],nk1,nk2,nk3,add1,add2,add3))

        HRs = None
        Hksp = np.reshape(Hksp, (Hksp.shape[0], nfft1 * nfft2 * nfft3))
        Hksp = gather_full(Hksp, npool)

        if rank == 0:
          Hksp = np.ascontiguousarray(Hksp.T)
          Hksp = np.reshape(Hksp, (nfft1 * nfft2 * nfft3, nawf, nawf))
        else:
          Hksp = np.zeros((nfft1 * nfft2 * nfft3, nawf, nawf), dtype=complex)

        comm.Bcast(Hksp)

        # Hksp=None
        # Hksp=np.zeros((HRs.shape[0],nfft1,nfft2,nfft3),dtype=complex)

        # for m in range(Hksp.shape[0]):
        #     Hksp[m,:,:,:]=np.fft.fftn(zero_pad(HRs[m,:,:,:],nk1,nk2,nk3,add1,add2,add3))

        # HRs  = None
        # Hksp = np.reshape(Hksp,(Hksp.shape[0],nfft1*nfft2*nfft3))
        # Hksp = gather_full(Hksp,npool)
        # if rank==0:
        #     Hksp = np.ascontiguousarray(Hksp.T)
        #     Hksp = np.reshape(Hksp,(nfft1*nfft2*nfft3,nawf,nawf))
        # else:
        #     Hksp=np.zeros((nfft1*nfft2*nfft3,nawf,nawf),dtype=complex)
        # comm.Bcast(Hksp)

      else:
        HRs = np.zeros((Hksp.shape[0], nfft1, nfft2, nfft3), dtype=complex)
        for m in range(Hksp.shape[0]):
          HRs[m, :, :, :] = zero_pad(Hksp[m, :, :, :], nk1, nk2, nk3, add1, add2, add3)
        Hksp = None
        Hksp = np.fft.fftn(HRs, axes=(1, 2, 3))
        HRs = None
        Hksp = np.reshape(Hksp, (Hksp.shape[0], nfft1 * nfft2 * nfft3))
        Hksp = gather_scatter(Hksp, 1, npool)
        Hksp = np.ascontiguousarray(Hksp.T)
        Hksp = gather_full(Hksp, npool)
        if rank == 0:
          Hksp = np.reshape(Hksp, (nfft1 * nfft2 * nfft3, nawf, nawf))
          print(Hksp[0, :8, :8].real)

        Hksp = np.reshape(Hksp, (nfft1 * nfft2 * nfft3, nawf, nawf))

      # if it's the non interpolated grid
      if i % 2:
        Hksp, _ = symmetrize_grid(Hksp, U, a_index, phase_shifts, inv_flag, U_inv,
                                  sym_TR, full_grid, jchia, spin_orb, mag_calc,
                                  nfft1, nfft2, nfft3, nkl_no_interp, partial_grid, npool)

      # if it's the interpolated grid
      else:
        Hksp, tmax = symmetrize_grid(Hksp, U, a_index, phase_shifts, inv_flag, U_inv,
                                     sym_TR, full_grid_interp, jchia, spin_orb,
                                     mag_calc, nfft1, nfft2, nfft3, nkl_interp, partial_grid_interp, npool)

      nk1 += add1
      nk2 += add2
      nk3 += add3

      if rank == 0 and i % 2 and verbose:
        print("Sym iter #%2d: %6.4e" % ((i // 2) + 1, tmax[0]))

      # stop if we hit threshold
      if tmax < thresh:
        if i % 2 and i >= 3:
          break

  # for debugging purposes
  # try:
  #    if rank==0:
  #        check(Hksp,si_per_k,new_k_ind,orig_k_ind,phase_shifts,U,
  #              a_index,inv_flag,equiv_atom,kp,symop,full_grid,sym_info,sym_TR)
  # except Exception as e:
  #    print(('Exception in pao_sym, check.', e), flush=True)

  return Hksp

def open_grid_nspin2(Hksp, full_grid, kp, symop, symop_cart, atom_pos, shells, a_index, equiv_atom,
              nk1, nk2, nk3, o1, o2, o3, sym_TR, symm_grid, thresh, max_iter, verbose, npool):
  # calculates full H(k) grid from wedge
  nawf = Hksp.shape[1]

  H0 = Hksp[...,0]
  H1 = Hksp[...,1]
  # get inversion operator
  U_inv = get_inv_op(shells)

  # apply time reversal symmetry H(k) = H(-k)* where appropriate
  H0, _ = apply_t_rev(H0, kp, False, U_inv, None)
  H1, kp = apply_t_rev(H1, kp, False, U_inv, None)

  # get array with wigner_d rotation matrix for each symop
  # for each of the orbital angular momentum l=[0,1,2,3]
  wigner, inv_flag = get_wigner(symop_cart)
  # convert the wigner_d into chemistry form for each symop
  wigner = convert_wigner_d(wigner)

  # get phase shifts from rotation symop
  phase_shifts = get_phase_shifts(atom_pos, symop, equiv_atom)

  # build U and U_inv from blocks
  U = build_U_matrix(wigner, shells)

  # adds transformation to U that maps orbitals from
  # atom A to equivalent atom B atoms after symop
  U_wyc = map_equiv_atoms(a_index, equiv_atom)

  # combine U_wyc and U
  U = add_U_wyc(U, U_wyc)

  # get index of k in wedge, index in full grid,
  # and index of symop that transforms k to k'
  new_k_ind, orig_k_ind, si_per_k = find_equiv_k(kp, symop, full_grid, sym_TR, check=True)


  # transform H(k) -> H(k')

  H0 = wedge_to_grid(H0, U, a_index, phase_shifts, kp,
                       new_k_ind, orig_k_ind, si_per_k, inv_flag, U_inv, sym_TR, False, npool)
  H1 = wedge_to_grid(H1, U, a_index, phase_shifts, kp,
                        new_k_ind, orig_k_ind, si_per_k, inv_flag, U_inv, sym_TR, False, npool)

  if rank == 0:
    # enforce time reversion where appropriate

    H0 = enforce_t_rev(H0, nk1, nk2, nk3, False, U_inv, None)
    H1 = enforce_t_rev(H1, nk1, nk2, nk3, False, U_inv, None)

    if np.any(sym_TR) and (not np.all(sym_TR)):
      # determine which mapping of k involves TR
      k_TR = sym_TR[si_per_k]
      Hksp_up = np.zeros_like(H0)
      Hksp_down = np.zeros_like(H1)
      Hksp_up[~k_TR] = H0[~k_TR]
      Hksp_up[k_TR] = H1[k_TR]
      Hksp_down[~k_TR] = H1[~k_TR]
      Hksp_down[k_TR] = H0[k_TR]
    else:
      Hksp_up = H0
      Hksp_down = H1

    Hksp = np.stack((Hksp_up,Hksp_down),axis=-1)

  else:
    Hksp = np.zeros((full_grid.shape[0], nawf, nawf,2), dtype=complex)

  comm.Bcast(Hksp)

  if symm_grid:

    symop_inv = np.zeros_like(symop)
    for i in range(symop.shape[0]):
      symop_inv[i] = LA.inv(symop[i])

    nkl = []
    partial_grid = scatter_full(full_grid, npool)
    for i in range(partial_grid.shape[0]):
      nkl.append(find_equiv_k(partial_grid[i][None], symop_inv, full_grid, sym_TR, check=False, include_self=True))
    nkl_no_interp = np.array(nkl)

    Hksp, tmax = symmetrize_grid_nspin2(Hksp, U, a_index, phase_shifts, inv_flag, U_inv, sym_TR,
                                   full_grid, nk1, nk2, nk3,
                                   nkl_no_interp, partial_grid, npool)
    upscale1 = int(0.25 * nk1)
    upscale2 = int(0.25 * nk2)
    upscale3 = int(0.25 * nk3)

    nfft1 = nk1 + upscale1
    nfft2 = nk2 + upscale2
    nfft3 = nk3 + upscale3

    full_grid_interp = get_full_grid(nfft1, nfft2, nfft3, o1, o2, o3)
    nkl = []
    partial_grid_interp = scatter_full(full_grid_interp, npool)
    for i in range(partial_grid_interp.shape[0]):
      nkl.append(
        find_equiv_k(partial_grid_interp[i][None], symop_inv, full_grid_interp, sym_TR, check=False, include_self=True))
    nkl_interp = np.array(nkl)
    # max difference bewtween H(k) and H(k*)
    tmax = [999999,999999]

    for i in range(int(max_iter * 2)):
      st = time.time()
      add1 = upscale1 * ((-1) ** i)
      add2 = upscale2 * ((-1) ** i)
      add3 = upscale3 * ((-1) ** i)

      nfft1 = nk1 + add1
      nfft2 = nk2 + add2
      nfft3 = nk3 + add3

      if rank == 0:
        Hksp = np.reshape(Hksp, (nk1 * nk2 * nk3, nawf * nawf, 2))
        Hksp = np.ascontiguousarray(np.transpose(Hksp,axes=(1,0,2)))

      Hksp = scatter_full(Hksp, npool)

      Hksp = np.reshape(Hksp, (Hksp.shape[0], nk1, nk2, nk3, 2))
      HRs = np.fft.ifftn(Hksp, axes=(1, 2, 3))

      Hksp = None
      Hksp = np.zeros((HRs.shape[0], nfft1, nfft2, nfft3, 2), dtype=complex)

      for m in range(Hksp.shape[0]):
        Hksp[m, :, :, :, 0] = np.fft.fftn(zero_pad(HRs[m, :, :, :, 0], nk1, nk2, nk3, add1, add2, add3))
        Hksp[m, :, :, :, 1] = np.fft.fftn(zero_pad(HRs[m, :, :, :, 1], nk1, nk2, nk3, add1, add2, add3))

      HRs = None
      Hksp = np.reshape(Hksp, (Hksp.shape[0], nfft1 * nfft2 * nfft3, 2))
      Hksp = gather_full(Hksp, npool)

      if rank == 0:
        Hksp = np.ascontiguousarray(np.transpose(Hksp,axes=(1,0,2)))
        Hksp = np.reshape(Hksp, (nfft1 * nfft2 * nfft3, nawf, nawf, 2))
      else:
        Hksp = np.zeros((nfft1 * nfft2 * nfft3, nawf, nawf, 2), dtype=complex)

      comm.Bcast(Hksp)

      # if it's the non interpolated grid
      if i % 2:
        Hksp, _ = symmetrize_grid_nspin2(Hksp, U, a_index, phase_shifts, inv_flag, U_inv,
                  sym_TR, full_grid, nfft1, nfft2, nfft3, nkl_no_interp, partial_grid, npool)


      # if it's the interpolated grid
      else:
        Hksp, tmax = symmetrize_grid_nspin2(Hksp, U, a_index, phase_shifts, inv_flag, U_inv,
                    sym_TR, full_grid_interp, nfft1, nfft2, nfft3, nkl_interp, partial_grid_interp, npool)


      nk1 += add1
      nk2 += add2
      nk3 += add3

      if rank == 0 and i % 2 and verbose:
        print("Sym iter #%2d: up %6.4e, down %6.4e" % ((i // 2) + 1, tmax[0], tmax[1]))

      # stop if we hit threshold
      if max(tmax) < thresh:
        if i % 2 and i >= 3:
          break

  return Hksp


############################################################################################
############################################################################################
############################################################################################

def open_grid_wrapper(data_controller):
  #    np.set_printoptions(precision=3,suppress=True,linewidth=220)

  # wrapper function to unload everything from
  # data controller and do a few conversions

  data_arrays = data_controller.data_arrays
  data_attr = data_controller.data_attributes
  alat = data_attr['alat']
  # nelec       = data_attr['nelec']
  nk1 = data_attr['nk1']
  nk2 = data_attr['nk2']
  nk3 = data_attr['nk3']
  o1 = data_attr['ok1']
  o2 = data_attr['ok2']
  o3 = data_attr['ok3']
  spin_orb = data_attr['dftSO']
  mag_calc = data_attr['dftMAG']
  symm_grid = data_attr['symmetrize']
  thresh = data_attr['symm_thresh']
  max_iter = data_attr['symm_max_iter']
  verbose = data_attr['verbose']
  Hks = data_arrays['Hks']
  atom_pos = data_arrays['tau'] / alat
  atom_lab = data_arrays['atoms']
  equiv_atom = data_arrays['equiv_atom']
  kp_red = data_arrays['kpnts']
  b_vectors = data_arrays['b_vectors']
  a_vectors = data_arrays['a_vectors']
  # sym_info    = data_arrays['sym_info']
  # sym_shift   = data_arrays['sym_shift']
  symop = data_arrays['sym_rot']
  sym_TR = data_arrays['sym_TR']
  npool = data_attr['npool']

  a_vectors = a_vectors
  b_vectors = b_vectors
  nspin = Hks.shape[3]
  nawf = Hks.shape[0]

  # convert atomic positions to crystal fractional coords
  conv = LA.inv(a_vectors)
  atom_pos = atom_pos @ conv
  atom_pos = correct_roundoff(atom_pos)
  atom_pos = np.around(atom_pos, decimals=6)

  # get symop crystal -> cartesian
  symop = correct_roundoff(symop)
  symop_cart = np.zeros_like(symop, dtype=float)
  inv_a_vectors = LA.inv(a_vectors)
  for isym in range(symop.shape[0]):
    symop_cart[isym] = (inv_a_vectors @ symop[isym] @ a_vectors)

  symop_cart = correct_roundoff(symop_cart, incl_hex=True, atol=1.e-6)

  # convert k points from cartesian to crystal fractional
  conv = LA.inv(b_vectors)
  conv = correct_roundoff(conv)
  kp_red = kp_red @ conv
  kp_red = correct_roundoff(kp_red)

  # get full grid in crystal fractional coords
  full_grid = get_full_grid(nk1, nk2, nk3, o1, o2, o3)

  jchia = []
  shells = []
  a_index = []
  sh = data_arrays['shells']
  for i, a in enumerate(atom_lab):
    ash = []
    for v in sh[a]:
      ash += [v, v] if v == 0 and spin_orb else [v]
    shells += ash
    a_index += [i] * np.sum([2 * n + 1 for n in ash])
    if spin_orb:
      jchia += data_arrays['jchia'][a]

  jchia = np.array(jchia)
  shells = np.array(shells)
  a_index = np.array(a_index)

  # correct small differences due to conversion
  kp_red = correct_roundoff_kp(kp_red, full_grid)

  # we wont need this for now
  if rank == 0:
    if nspin == 2:
      data_arrays['Hks'] = np.zeros((nawf, nawf, nk1 * nk2 * nk3, nspin), dtype=complex)
    else:
      data_arrays['Hks'] = None
  else:
    data_arrays['Hks'] = None

  # expand grid from wedge (older version that works for QE)
  # for ispin in range(nspin):
  #     Hksp = np.ascontiguousarray(np.transpose(Hks, axes=(2, 0, 1, 3))[:, :, :, ispin])
  #
  #     Hksp = open_grid(Hksp, full_grid, kp_red, symop, symop_cart, atom_pos,
  #                      shells, a_index, equiv_atom, sym_info, sym_shift,
  #                      nk1, nk2, nk3, spin_orb, sym_TR, jchia, mag_calc,
  #                      symm_grid, thresh, max_iter, nelec, verbose)
  #
  #     if rank == 0:
  #         if nspin == 2:
  #             data_arrays['Hks'][:, :, :, ispin] = np.ascontiguousarray(np.transpose(Hksp, axes=(1, 2, 0)))
  #         else:
  #             data_arrays['Hks'] = np.ascontiguousarray(np.transpose(Hksp, axes=(1, 2, 0))[..., None])
  #     #            np.save("kham.npy",np.ascontiguousarray(np.transpose(Hksp,axes=(1,2,0))))
  #     else:
  #         Hksp = None

  Hksp = np.ascontiguousarray(np.transpose(Hks, axes=(2, 0, 1, 3)))
  if nspin == 1:
    Hks_full = open_grid(Hksp[...,0], full_grid, kp_red, symop, symop_cart, atom_pos,
                     shells, a_index, equiv_atom,
                     nk1, nk2, nk3, o1, o2, o3, spin_orb, sym_TR, jchia, mag_calc,
                     symm_grid, thresh, max_iter, verbose, npool)
    Hks_full = Hks_full[...,np.newaxis]
  else:
    Hks_full = open_grid_nspin2(Hksp, full_grid, kp_red, symop, symop_cart, atom_pos,
                         shells, a_index, equiv_atom,
                         nk1, nk2, nk3, o1, o2, o3, sym_TR,
                         symm_grid, thresh, max_iter, verbose, npool)
  if rank == 0:
    # if nspin==1:
    #     data_arrays['Hks'] = np.ascontiguousarray(np.transpose(Hksp, axes=(1, 2, 0))[..., None])
    #     #  np.save("kham.npy",np.ascontiguousarray(np.transpose(Hksp,axes=(1,2,0))))
    # else:
    #     data_arrays['Hks'][:, :, :, 0] = np.ascontiguousarray(np.transpose(Hksp_up, axes=(1, 2, 0)))
    #     data_arrays['Hks'][:, :, :, 1] = np.ascontiguousarray(np.transpose(Hksp_down, axes=(1, 2, 0)))
    data_arrays['Hks'] = np.ascontiguousarray(np.transpose(Hks_full, axes=(1, 2, 0, 3)))
  # else:
  #   Hks_full = None


############################################################################################
############################################################################################
############################################################################################

def symmetrize(Hksp, U, a_index, phase_shifts, new_k_ind, orig_k_ind, si_per_k, inv_flag, U_inv, sym_TR, spin_orb,
               full_grid, reverse=False):
  # generates full grid from k points in IBZ
  nawf = Hksp.shape[1]
  Hksp_s = np.zeros((new_k_ind.shape[0], nawf, nawf), dtype=complex)

  for j in range(new_k_ind.shape[0]):
    isym = si_per_k[j]
    oki = orig_k_ind[j]
    nki = new_k_ind[j]

    # if symop is identity
    if isym == 0:
      Hksp_s[j] = Hksp[nki]
      continue

    # other cases
    H = Hksp[nki]

    # get k dependent U
    U_k = get_U_k(full_grid[nki], phase_shifts[isym], a_index, U[isym])

    # transformated H(k)
    if not reverse:
      THP = U_k @ H @ np.conj(U_k.T)
    else:
      THP = np.conj(U_k.T) @ H @ U_k

    # apply inversion operator if needed
    if inv_flag[isym]:
      THP *= U_inv

    # time inversion is anti-unitary
    # old version
    # if sym_TR[isym]:
    #    THP *= U_inv
    #    THP = np.conj(THP)
    if sym_TR[isym]:
      if spin_orb:
        THP *= U_inv
      THP = np.conj(THP)

    Hksp_s[j] = THP

  return Hksp_s


############################################################################################
############################################################################################
############################################################################################

def symmetrize_grid(Hksp, U, a_index, phase_shifts, inv_flag, U_inv, sym_TR, full_grid, jchia, spin_orb,
                    mag_calc, nk1, nk2, nk3, nkl, partial_grid, npool):
  max_iter = 1
  tmax = []
  Hksp_d = np.zeros((partial_grid.shape[0], Hksp.shape[1], Hksp.shape[2]), dtype=complex)

  for i in range(partial_grid.shape[0]):
    new_k_ind, orig_k_ind, si_per_k = nkl[i]

    temp = symmetrize(Hksp, U, a_index, phase_shifts, new_k_ind,
                      orig_k_ind, si_per_k, inv_flag, U_inv, sym_TR, spin_orb, full_grid)

    tmax.append(np.amax(np.abs(temp[0][None] - temp)))
    Hksp_d[i] = np.sum(temp, axis=0) / (temp.shape[0])

  # make sure of hermiticity of each H(k)
  Hksp_d = enforce_hermaticity(Hksp_d)

  if rank == 0:
    Hksp = gather_full(Hksp_d, npool)
    if not (spin_orb and mag_calc):
      Hksp = enforce_t_rev(Hksp, nk1, nk2, nk3, spin_orb, U_inv, jchia)

  else:
    gather_full(Hksp_d, npool)

  tm = np.array([np.amax(tmax)])
  tmax = np.copy(tm)
  comm.Reduce(tm, tmax, op=MPI.MAX)
  comm.Bcast(tmax)

  return Hksp, tmax

def symmetrize_grid_nspin2(Hksp, U, a_index, phase_shifts, inv_flag, U_inv, sym_TR, full_grid,
                     nk1, nk2, nk3, nkl, partial_grid, npool):
  nawf = Hksp.shape[1]

  tmax_up = []; tmax_down = []
  H0 = Hksp[..., 0]
  H1 = Hksp[..., 1]
  Hksp_d_up = np.zeros((partial_grid.shape[0], nawf, nawf), dtype=complex)
  Hksp_d_down = np.zeros((partial_grid.shape[0], nawf, nawf), dtype=complex)
  for i in range(partial_grid.shape[0]):
    new_k_ind, orig_k_ind, si_per_k = nkl[i]
    temp0 = symmetrize(H0, U, a_index, phase_shifts, new_k_ind,
                      orig_k_ind, si_per_k, inv_flag, U_inv, sym_TR, False, full_grid)
    temp1 = symmetrize(H1, U, a_index, phase_shifts, new_k_ind,
                       orig_k_ind, si_per_k, inv_flag, U_inv, sym_TR, False, full_grid)
    if np.any(sym_TR) and (not np.all(sym_TR)):
      temp_up = np.zeros_like(temp0)
      temp_down = np.zeros_like(temp1)
      temp_up[~sym_TR] = temp0[~sym_TR]
      temp_up[sym_TR] = temp1[sym_TR]
      temp_down[~sym_TR] = temp1[~sym_TR]
      temp_down[sym_TR] = temp0[sym_TR]
    else:
      temp_up = temp0
      temp_down = temp1

    tmax_up.append(np.amax(np.abs(temp_up[0][None] - temp_up)))
    tmax_down.append(np.amax(np.abs(temp_down[0][None] - temp_down)))
    Hksp_d_up[i] = np.sum(temp_up, axis=0) / (temp_up.shape[0])
    Hksp_d_down[i] = np.sum(temp_down, axis=0) / (temp_down.shape[0])

  # make sure of hermiticity of each H(k)
  Hksp_d_up = enforce_hermaticity(Hksp_d_up)
  Hksp_d_down = enforce_hermaticity(Hksp_d_down)


  if rank == 0:

    Hksp_up = gather_full(Hksp_d_up, npool)
    Hksp_down = gather_full(Hksp_d_down, npool)

    Hksp_up = enforce_t_rev(Hksp_up, nk1, nk2, nk3, False, U_inv, None)
    Hksp_down = enforce_t_rev(Hksp_down, nk1, nk2, nk3, False, U_inv, None)

    Hksp = np.stack((Hksp_up,Hksp_down),axis=-1)

  else:
    gather_full(Hksp_d_up, npool)
    gather_full(Hksp_d_down, npool)


  tm_up = np.array([np.amax(tmax_up)])
  tm_down = np.array([np.amax(tmax_down)])
  tmax_up = np.copy(tm_up)
  tmax_down = np.copy(tm_down)
  comm.Reduce(tm_up, tmax_up, op=MPI.MAX)
  comm.Reduce(tm_down, tmax_down, op=MPI.MAX)
  comm.Bcast(tmax_up)
  comm.Bcast(tmax_down)
  tmax = [tmax_up[0],tmax_down[0]]

  return Hksp, tmax


############################################################################################
############################################################################################
############################################################################################

def correct_roundoff_kp(kp, full_grid):
  kp_c = np.copy(kp)
  nw = np.where(np.isclose(cdist(kp_c, full_grid), 0.0, atol=1.e-5, rtol=1.e-5, ))
  kp_c[nw[0]] = full_grid[nw[1]]

  return kp_c

############################################################################################
############################################################################################
############################################################################################



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

import numpy as np

from mpi4py import MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()

# Block diagonalization of Hermitian matrices
'''L S Cederbaum, J Schirmer and H-D Meyer
J. Phys. A: Mat. Gen. 22, 2427 (1989)

The formula for the block diagonal part of the S matrix is from:
B.N. Parlett, The symmetric eigenvalue problem (SIAM's Classics in Applied Mathematics, 1998)
pag. 45, sec. 3.1.1 remark n. 5.'''

def do_minimal(data_controller,first_band):
    
    # this eliminates bands at the bottom

    arry,attr = data_controller.data_dicts()
    
    import numpy.random as rd
    from numpy import linalg as npl
    
    # bnd = attr['bnd']
    bnd = first_band
    nawf = attr['nawf']
    nspin = attr['nspin']
    nkpnts = attr['nkpnts']
    basis = arry['basis']
    
    # arry['Hksp'] = np.reshape(arry['Hksp'],(nawf,nawf,nkpnts,nspin))
    
    Hks = np.zeros((nkpnts,nawf-bnd,nawf-bnd,nspin), dtype=complex)
    for ik in range(nkpnts):
        for ispin in range(nspin):
            Sbd = np.zeros((nawf,nawf),dtype=complex)
            Sbdi = np.zeros((nawf,nawf),dtype=complex)
            S = sv = np.zeros((nawf,nawf),dtype=complex)
            e = se = np.zeros(nawf,dtype=float)
            e,S = npl.eigh(arry['Hksp'][ik,:,:,ispin])
            S11 = S[:bnd,:bnd] + 1.0*rd.random(bnd)/10000.
            S21 = S[:bnd,bnd:] + 1.0*rd.random(nawf-bnd)/10000.
            S12 = S21.T
            S22 = S[bnd:,bnd:] + 1.0*rd.random(nawf-bnd)/10000.
            S22 = S22 + S21.T.dot(np.dot(npl.inv(S11),S12.T))
            Sbd[:bnd,:bnd] = 0.5*(S11+np.conj(S11.T))
            Sbd[bnd:,bnd:] = 0.5*(S22+np.conj(S22.T))
            Sbdi = npl.inv(np.dot(Sbd,np.conj(Sbd.T)))
            se,sv = npl.eigh(Sbdi)
            se = np.sqrt(se+0.0j)*np.identity(nawf,dtype=complex)
            Sbdi = sv.dot(se).dot(np.conj(sv).T)
            T = S.dot(np.conj(Sbd.T)).dot(Sbdi)
            Hbd = np.conj(T.T).dot(np.dot(arry['Hksp'][ik,:,:,ispin],T))
            # Hks[:,:,ik,ispin] = 0.5*(Hbd[:bnd,:bnd]+np.conj(Hbd[:bnd,:bnd].T))
            # Hks[:,:,ik,ispin] = 0.5*(Hbd[first_band:,first_band:]+np.conj(Hbd[first_band:,first_band:].T))
            Hks[ik,:,:,ispin] = Hbd[first_band:,first_band:]
            
    arry['Hksp'] = Hks #np.reshape(Hks,(nkpnts,nawf-bnd,nawf-bnd,nspin))
    attr['nawf'] = nawf-bnd
    # ashape = (attr['nawf'],attr['nawf'],attr['nk1'],attr['nk2'],attr['nk3'],attr['nspin'])
    # arry['Hks'] = np.reshape(arry['Hks'], ashape)
    arry['Dnm'] = np.empty((attr['nawf'],attr['nawf'],3))
    for i in range(3):
      for n in range(attr['nawf']):
        for m in range(attr['nawf']):
          arry['Dnm'][n,m,i] = basis[n+first_band]['tau'][i] - basis[m+first_band]['tau'][i]



#
# PAOFLOW
#
# Copyright 2016-2022 - Marco BUONGIORNO NARDELLI (mbn@unt.edu)
#
# Reference:
#
# F.T. Cerasoli, A.R. Supka, A. Jayaraj, I. Siloi, M. Costa, J. Slawinska, S. Curtarolo, M. Fornari, D. Ceresoli, and M. Buongiorno Nardelli,
# Advanced modeling of materials with PAOFLOW 2.0: New features and software design, Comp. Mat. Sci. 200, 110828 (2021).
#
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang, 
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on 
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .

from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def do_rashba_edelstein (data_controller, ene, temperature, regularization, twoD_structure, lattice_height, structure_thickness, write_to_file):
  import numpy as np
  from os.path import join
  from .smearing import gaussian
  from .constants import ELECTRONVOLT_SI,BOHR_RADIUS_CM,HBAR,LL

  comm,rank = data_controller.comm,data_controller.rank
  arrays,attr = data_controller.data_dicts()

  snktot = arrays['v_k'].shape[0]
  ind_plot = arrays['ind_plot']
  nstates = len(ind_plot)
  nktot = attr['nkpnts']
  tau_const = 1.
  esize = ene.size

  pksp = np.take(np.diagonal(np.real(arrays['pksp'][:,:,:,:,0]),axis1=2,axis2=3), ind_plot, axis=2)

  deltakp = np.take(arrays['deltakp'], ind_plot, axis=1)[:,:,0]
  E_k = np.take(arrays['E_k'], ind_plot, axis=1)[:,:,0]
  St = np.real(arrays['sktxt'])

  kai_aux = np.zeros((snktot,3,3,nstates), dtype=float)
  j_aux = np.zeros((snktot,3,3,nstates), dtype=float)
  for l in range(3):
    for m in range(3):
      kai_aux[:,l,m,:] = tau_const*St[:,l,:]*pksp[:,m,:]
      j_aux[:,l,l,:] = tau_const*pksp[:,l,:]*pksp[:,l,:]   

  kai_eaux = np.zeros((snktot,3,3,esize), dtype=float)
  j_eaux = np.zeros((snktot,3,3,esize), dtype=float)

  def dfermi(E,ene,temp):
    return -1/(4*temp*(np.cosh((E-ene)/(2*temp))**2))

  for i in range(esize):
    gaussian_smear = None
    if attr['smearing'] == 'gauss':
      if temperature == 0:
        gaussian_smear = gaussian(E_k, ene[i], deltakp)
      else:
        gaussian_smear = dfermi(E_k, ene[i], temperature)
    else:
      raise ValueError('Routine requires \'gauss\' smearing')
    for l in range(3):
      for m in range(3):
        kai_eaux[:,l,m,i] = np.sum(kai_aux[:,l,m,:]*gaussian_smear, axis=1)
        j_eaux[:,l,l,i] = np.sum(j_aux[:,l,l,:]*gaussian_smear, axis=1)
  kai_aux = None
  j_aux = None

  kai = (np.zeros((3,3,esize),dtype=float) if rank==0 else None)
  jc = (np.zeros((3,3,esize),dtype=float) if rank==0 else None)

  kai_eaux = np.ascontiguousarray(np.sum(kai_eaux,axis=0))
  j_eaux = np.ascontiguousarray(np.sum(j_eaux,axis=0))

  comm.Reduce(kai_eaux, kai, op=MPI.SUM)
  comm.Reduce(j_eaux, jc, op=MPI.SUM)

  if rank == 0:
    Ekai = np.empty((3,3,esize), dtype=float)
    for i in range(3):
      for j in range(3):
        Ekai[i,j] = -HBAR * kai[i,j] / (jc[j,j] * ELECTRONVOLT_SI * BOHR_RADIUS_CM + regularization)

    if twoD_structure:
      Ekai *= (lattice_height / structure_thickness)

    sEkai = { 0:'x', 1:'y', 2:'z' }
    wEkai = lambda fn,e,t : fn.write('% .5f % 9.5e\n'%(e,t))
    wtup = lambda fn,tu : fn.write('% .5f % 9.5e % 9.5e % 9.5e % 9.5e % 9.5e % 9.5e % 9.5e % 9.5e % 9.5e\n'%tu)
    gtup = lambda tu,i : (ene[i],tu[0,0,i],tu[0,1,i],tu[0,2,i],tu[1,0,i],tu[1,1,i],tu[1,2,i],tu[2,0,i],tu[2,1,i],tu[2,2,i])

    if write_to_file:
      fkai = open(join(attr['opath'],'kai.dat'), 'w')
      fcurrent = open(join(attr['opath'],'current.dat'), 'w')

      ofE = lambda si,sj : open(join(attr['opath'],f'Ekai_{si}{sj}.dat'), 'w')
      fEkai = [[ofE(sEkai[i],sEkai[j]) for j in range(3)] for i in range(3)]

      for ie in range(esize):
        wtup(fkai,gtup(kai,ie))
        wtup(fcurrent,gtup(jc,ie))
        for i in range(3):
          for j in range(3):
            wEkai(fEkai[i][j], ene[ie], Ekai[i,j,ie])

      fkai.close()
      fcurrent.close()
      for i in range(3):
        for j in range(3):
          fEkai[i][j].close()




from matplotlib import pyplot as plt

def plot_dos ( es, dos, title, x_lim, y_lim, vertical, col, x_label=None, y_label=None ):
  '''
  '''

  fig = plt.figure()

  tit = 'DoS' if title is None else title
  fig.suptitle(tit)

  ax = fig.add_subplot(111)

  if vertical:
    ax.plot(dos, es, color=col)
  else:
    ax.plot(es, dos, color=col)
  if not x_lim is None:
    ax.set_xlim(*x_lim)
  elif vertical:
    ax.set_xlim(0, ax.get_xlim()[1])
  if not y_lim is None:
    ax.set_ylim(*y_lim)
  elif not vertical:
    ax.set_ylim(0, ax.get_ylim()[1])

  el = 'Energy (eV)' if x_label is None else x_label
  dl = 'electrons/eV' if y_label is None else y_label
  xl = el if not vertical else dl
  yl = dl if not vertical else el

  ax.set_xlabel(xl, fontsize=12)
  ax.set_ylabel(yl, fontsize=12)

  plt.show()


def plot_pdos ( es, dos, title, x_lim, y_lim, vertical, cols, labels, legend ):
  '''
  '''
  import numpy as np

  if labels is None:
    labels = list(range(len(dos)))
  else:
    if len(labels) != len(dos):
      raise Exception('Must provide one label for each pdos file')

  if cols is None or isinstance(cols, str):
    cols = [cols] * len(dos)
  else:
    cols = np.array(cols)
    cs = cols.shape
    if len(cs) == 1:
      cols = [cols] * len(dos)
    elif cs[0] != len(labels):
      raise Exception('Must provide one color for each pdos file')

  fig = plt.figure()

  tit = 'PDoS' if title is None else title
  fig.suptitle(tit)

  ax = fig.add_subplot(111)

  if vertical:
    for i,d in enumerate(dos):
      ax.plot(d, es, color=cols[i], label=labels[i])
  else:
    for i,d in enumerate(dos):
      ax.plot(es, d, color=cols[i], label=labels[i])
  if not x_lim is None:
    ax.set_xlim(*x_lim)
  elif vertical:
    ax.set_xlim(0, ax.get_xlim()[1])
  if not y_lim is None:
    ax.set_ylim(*y_lim)
  elif not vertical:
    ax.set_ylim(0, ax.get_ylim()[1])

  el = 'Energy (eV)'
  dl = 'electrons/eV'
  xl = el if not vertical else dl
  yl = dl if not vertical else el

  ax.set_xlabel(xl, fontsize=12)
  ax.set_ylabel(yl, fontsize=12)

  if legend:
    ax.legend()

  plt.show()


def plot_bands ( bands, sym_points, title, label, y_lim, col ):
  '''
  '''

  fig = plt.figure()

  tit = 'Band Structure' if title is None else title
  fig.suptitle(tit)

  ax = fig.add_subplot(111)

  for b in bands:
    ax.plot(b, color=col)
  if y_lim is None:
    y_lim = ax.get_ylim()
  ax.set_xlim(0, bands.shape[1])
  ax.set_ylim(*y_lim)
  if sym_points is None:
    ax.xaxis.set_visible(False)
  else:
    ax.set_xticks(sym_points[0])
    ax.set_xticklabels(sym_points[1])
    ax.vlines(sym_points[0], y_lim[0], y_lim[1], color='gray')
  if label is None:
    label = '$\epsilon$($\\bf{k}$) (eV)'
  ax.set_ylabel(label, fontsize=12)

  plt.show()


def plot_dos_beside_bands ( es, dos, bands, sym_points, title, band_label, x_lim, y_lim, col, dos_ticks ):
  '''
  '''
  from matplotlib import gridspec

  fig = plt.figure()
  spec = gridspec.GridSpec(ncols=2, nrows=1, width_ratios=[5,1])

  tit = 'Band Structure and DoS' if title is None else title
  fig.suptitle(tit)

  ax_b = fig.add_subplot(spec[0])
  ax_d = fig.add_subplot(spec[1])

  for b in bands:
    ax_b.plot(b, color=col)
  if y_lim is None:
    y_lim = ax_b.get_ylim()
  ax_b.set_xlim(0, bands.shape[1]-1)
  ax_b.set_ylim(*y_lim)
  if sym_points is None:
    ax_b.xaxis.set_visible(False)
  else:
    ax_b.set_xticks(sym_points[0])
    ax_b.set_xticklabels(sym_points[1])
    ax_b.vlines(sym_points[0], y_lim[0], y_lim[1], color='gray')
  if band_label is None:
    band_label = '$\epsilon$($\\bf{k}$) (eV)'
  ax_b.set_ylabel(band_label, fontsize=12)
  
  ax_d.plot(dos, es, color=col)
  if not x_lim is None:
    ax_d.set_xlim(*x_lim)
  else:
    ax_d.set_xlim(0, ax_d.get_xlim()[1])
  if not y_lim is None:
    ax_d.set_ylim(*y_lim)
  if not dos_ticks:
    ax_d.yaxis.set_visible(False)
    ax_d.xaxis.set_visible(False)
    plt.tight_layout()

  plt.show()


def plot_berry_under_bands ( berry, bands, sym_points, title, band_label, berry_label, x_lim, y_lim, col, dos_ticks ):
  '''
  '''
  from matplotlib import gridspec

  fig = plt.figure()
  spec = gridspec.GridSpec(ncols=1, nrows=2, height_ratios=[3,1])

  tit = 'Band Structure and Berry Phase' if title is None else title
  fig.suptitle(tit)

  ax_ba = fig.add_subplot(spec[0])
  ax_be = fig.add_subplot(spec[1])

  ax_be.plot(berry, color=col)
  for b in bands:
    ax_ba.plot(b, color=col)
  if y_lim is None:
    y_lim = ax_ba.get_ylim()
  ax_be.set_xlim(0, bands.shape[1]-1)
  ax_ba.set_xlim(0, bands.shape[1]-1)
  ax_ba.set_ylim(*y_lim)
  if sym_points is None:
    ax_be.xaxis.set_visible(False)
    ax_ba.xaxis.set_visible(False)
  else:
    tlim = ax_be.get_ylim()
    ax_be.set_ylim(*tlim)
    ax_be.set_xticks(sym_points[0])
    ax_be.set_xticklabels(sym_points[1])
    ax_be.vlines(sym_points[0], tlim[0], tlim[1], color='gray')
    ax_ba.set_xticks(sym_points[0])
    ax_ba.set_xticklabels(sym_points[1])
    ax_ba.vlines(sym_points[0], y_lim[0], y_lim[1], color='gray')

  if berry_label is None:
    berry_label = '$\Omega$($\\bf{k}$)'
  if band_label is None:
    band_label = '$\epsilon$($\\bf{k}$) (eV)'
  ax_be.set_ylabel(berry_label, fontsize=12)
  ax_ba.set_ylabel(band_label, fontsize=12)

  plt.show()
  quit()
  ax_d.plot(dos, es, color=col)
  if not x_lim is None:
    ax_d.set_xlim(*x_lim)
  else:
    ax_d.set_xlim(0, ax_d.get_xlim()[1])
  if not y_lim is None:
    ax_d.set_ylim(*y_lim)
  if not dos_ticks:
    ax_d.yaxis.set_visible(False)
    plt.tight_layout()

  plt.show()


def plot_tensor ( enes, tensors, eles, title, x_lim, y_lim, x_lab, y_lab, col, legend, min_zero=False ):
  '''
  '''
  import numpy as np

  fig = plt.figure()

  if title is None:
    raise ValueError('\'title\' cannot be None in plot_tensor')
  fig.suptitle(title)

  ax = fig.add_subplot(111)

  lmap = {0:'x', 1:'y', 2:'z'}
  lkey = lambda a,b : lmap[a] + lmap[b]
  if len(eles) == 0:
    tval = np.empty(tensors.shape[0], dtype=float)
    for i,v in enumerate(tensors):
      tval[i] = np.sum([v[j,j] for j in range(3)])/3
    col = col if type(col) is str else col[0]
    ax.plot(enes, tval, color=col, label='Avg.')
  else:
    if type(col) is str:
      for e in eles:
        ax.plot(enes, tensors[:,e[0],e[1]], color=col, label=lkey(*e))
    elif len(col) >= len(eles):
      for i,e in enumerate(eles):
        ax.plot(enes, tensors[:,e[0],e[1]], color=col[i], label=lkey(*e))
    else:
      for e in eles:
        ax.plot(enes, tensors[:,e[0],e[1]], label=lkey(*e))

  if not x_lim is None:
    ax.set_xlim(*x_lim)
  if not y_lim is None:
    ax.set_ylim(*y_lim)
  elif min_zero:
    ax.set_ylim(0, ax.get_ylim()[1])

  ax.set_xlabel(x_lab)
  ax.set_ylabel(y_lab)

  if legend:
    ax.legend()

  plt.show()


def plot_shc_tensor ( enes, shc, title, x_lim, y_lim, x_lab, y_lab, cols, labels, legend ):
  '''
  '''
  import numpy as np

  fig = plt.figure()

  if title is None:
    raise ValueError('\'title\' cannot be None in plot_tensor')
  fig.suptitle(title)

  ax = fig.add_subplot(111)

  if len(cols) >= len(shc):
    for i,s in enumerate(shc):
      ax.plot(enes, s, color=cols[i], label=labels[i])
  else:
    raise Exception('Dimensions of colors are incorrect. Blame GPAO.py')

  if not x_lim is None:
    ax.set_xlim(*x_lim)
  if not y_lim is None:
    ax.set_ylim(*y_lim)

  ax.set_xlabel(x_lab)
  ax.set_ylabel(y_lab)

  if legend:
    ax.legend()

  plt.show()

